@@ -4,17 +4,13 @@ use crate::{
     prelude::*,
     utils::{get_iter_capacity, NoNull},
 };
-pub use arrow::alloc;
-use arrow::array::{ArrayData, PrimitiveArray, PrimitiveBuilder};
-use arrow::array::{ArrayRef, LargeListBuilder};
-use arrow::{array::Array, buffer::Buffer};
-use num::Num;
-use polars_arrow::prelude::*;
+use arrow::{array::*, bitmap::Bitmap};
 use std::borrow::Cow;
 use std::iter::FromIterator;
 use std::marker::PhantomData;
 use std::sync::Arc;
 
+// N: the value type; T: the sentinel type
 pub trait ChunkedBuilder<N, T> {
     fn append_value(&mut self, val: N);
     fn append_null(&mut self);
@@ -28,7 +24,7 @@ pub trait ChunkedBuilder<N, T> {
 }
 
 pub struct BooleanChunkedBuilder {
-    array_builder: BooleanArrayBuilder,
+    array_builder: MutableBooleanArray,
     field: Field,
 }
 
@@ -36,17 +32,18 @@ impl ChunkedBuilder<bool, BooleanType> for BooleanChunkedBuilder {
     /// Appends a value of type `T` into the builder
     #[inline]
     fn append_value(&mut self, v: bool) {
-        self.array_builder.append_value(v);
+        self.array_builder.push(Some(v));
     }
 
     /// Appends a null slot into the builder
     #[inline]
     fn append_null(&mut self) {
-        self.array_builder.append_null();
+        self.array_builder.push(None);
     }
 
-    fn finish(mut self) -> BooleanChunked {
-        let arr = Arc::new(self.array_builder.finish());
+    fn finish(self) -> BooleanChunked {
+        let arr: BooleanArray = self.array_builder.into();
+        let arr = Arc::new(arr) as ArrayRef;
 
         ChunkedArray {
             field: Arc::new(self.field),
@@ -61,7 +58,7 @@ impl ChunkedBuilder<bool, BooleanType> for BooleanChunkedBuilder {
 impl BooleanChunkedBuilder {
     pub fn new(name: &str, capacity: usize) -> Self {
         BooleanChunkedBuilder {
-            array_builder: BooleanArrayBuilder::new(capacity),
+            array_builder: MutableBooleanArray::with_capacity(capacity),
             field: Field::new(name, DataType::Boolean),
         }
     }
@@ -72,7 +69,7 @@ where
     T: PolarsPrimitiveType,
     T::Native: Default,
 {
-    array_builder: PrimitiveBuilder<T>,
+    array_builder: MutablePrimitiveArray<T::Native>,
     field: Field,
 }
 
@@ -84,21 +81,19 @@ where
     /// Appends a value of type `T` into the builder
     #[inline]
     fn append_value(&mut self, v: T::Native) {
-        self.array_builder.append_value(v).unwrap()
+        self.array_builder.push(Some(v))
     }
 
     /// Appends a null slot into the builder
     #[inline]
     fn append_null(&mut self) {
-        self.array_builder.append_null().unwrap()
+        self.array_builder.push(None)
     }
 
-    fn finish(mut self) -> ChunkedArray<T> {
-        let arr = Arc::new(self.array_builder.finish());
-
+    fn finish(self) -> ChunkedArray<T> {
         ChunkedArray {
             field: Arc::new(self.field),
-            chunks: vec![arr],
+            chunks: vec![self.array_builder.into_arc()],
             phantom: PhantomData,
             categorical_map: None,
             ..Default::default()
@@ -111,15 +106,18 @@ where
     T: PolarsPrimitiveType,
 {
     pub fn new(name: &str, capacity: usize) -> Self {
+        let array_builder = MutablePrimitiveArray::<T::Native>::with_capacity(capacity)
+            .to(T::get_dtype().to_arrow());
+
         PrimitiveChunkedBuilder {
-            array_builder: PrimitiveBuilder::<T>::new(capacity),
+            array_builder,
             field: Field::new(name, T::get_dtype()),
         }
     }
 }
 
 pub struct Utf8ChunkedBuilder {
-    pub builder: LargeStringBuilder,
+    pub builder: MutableUtf8Array<i64>,
     pub capacity: usize,
     field: Field,
 }
@@ -133,7 +131,7 @@ impl Utf8ChunkedBuilder {
     /// * `bytes_capacity` - Number of bytes needed to store the string values.
     pub fn new(name: &str, capacity: usize, bytes_capacity: usize) -> Self {
         Utf8ChunkedBuilder {
-            builder: LargeStringBuilder::with_capacity(bytes_capacity, capacity),
+            builder: MutableUtf8Array::<i64>::with_capacities(capacity, bytes_capacity),
             capacity,
             field: Field::new(name, DataType::Utf8),
         }
@@ -142,25 +140,22 @@ impl Utf8ChunkedBuilder {
     /// Appends a value of type `T` into the builder
     #[inline]
     pub fn append_value<S: AsRef<str>>(&mut self, v: S) {
-        self.builder.append_value(v.as_ref()).unwrap();
+        self.builder.push(Some(v.as_ref()));
     }
 
     /// Appends a null slot into the builder
     #[inline]
     pub fn append_null(&mut self) {
-        self.builder.append_null().unwrap();
+        self.builder.push::<&str>(None);
     }
 
     #[inline]
     pub fn append_option<S: AsRef<str>>(&mut self, opt: Option<S>) {
-        match opt {
-            Some(s) => self.append_value(s.as_ref()),
-            None => self.append_null(),
-        }
+        self.builder.push(opt);
     }
 
-    pub fn finish(mut self) -> Utf8Chunked {
-        let arr = Arc::new(self.builder.finish());
+    pub fn finish(self) -> Utf8Chunked {
+        let arr = self.builder.into_arc();
         ChunkedArray {
             field: Arc::new(self.field),
             chunks: vec![arr],
@@ -199,23 +194,16 @@ impl ChunkedBuilder<Cow<'_, str>, Utf8Type> for Utf8ChunkedBuilderCow {
     }
 }
 
-/// Get the null count and the null bitmap of the arrow array
-pub fn get_bitmap<T: Array + ?Sized>(arr: &T) -> (usize, Option<&Buffer>) {
-    let data = arr.data();
-    (data.null_count(), data.null_buffer())
-}
-
 // Used in polars/src/chunked_array/apply.rs:24 to collect from aligned vecs and null bitmaps
-impl<T> FromIterator<(AlignedVec<T::Native>, Option<Buffer>)> for ChunkedArray<T>
+impl<T> FromIterator<(AlignedVec<T::Native>, Option<Bitmap>)> for ChunkedArray<T>
 where
     T: PolarsNumericType,
 {
-    fn from_iter<I: IntoIterator<Item = (AlignedVec<T::Native>, Option<Buffer>)>>(iter: I) -> Self {
+    fn from_iter<I: IntoIterator<Item = (AlignedVec<T::Native>, Option<Bitmap>)>>(iter: I) -> Self {
         let mut chunks = vec![];
 
         for (values, opt_buffer) in iter {
-            let arr = values.into_primitive_array::<T>(opt_buffer);
-            chunks.push(Arc::new(arr) as ArrayRef)
+            chunks.push(to_array::<T>(values, opt_buffer))
         }
         ChunkedArray::new_from_chunks("from_iter", chunks)
     }
@@ -237,11 +225,8 @@ where
     T: PolarsPrimitiveType,
 {
     fn new_from_slice(name: &str, v: &[T::Native]) -> Self {
-        let array_data = ArrayData::builder(T::DATA_TYPE)
-            .len(v.len())
-            .add_buffer(Buffer::from_slice_ref(&v))
-            .build();
-        ChunkedArray::new_from_chunks(name, vec![Arc::new(PrimitiveArray::<T>::from(array_data))])
+        let arr = PrimitiveArray::<T::Native>::from_slice(v).to(T::get_dtype().to_arrow());
+        ChunkedArray::new_from_chunks(name, vec![Arc::new(arr)])
     }
 
     fn new_from_opt_slice(name: &str, opt_v: &[Option<T::Native>]) -> Self {
@@ -299,16 +284,16 @@ where
     fn new_from_slice(name: &str, v: &[S]) -> Self {
         let values_size = v.iter().fold(0, |acc, s| acc + s.as_ref().len());
 
-        let mut builder = LargeStringBuilder::with_capacity(values_size, v.len());
+        let mut builder = MutableUtf8Array::<i64>::with_capacities(v.len(), values_size);
         v.iter().for_each(|val| {
-            builder.append_value(val.as_ref()).unwrap();
+            builder.push(Some(val.as_ref()));
         });
 
         let field = Arc::new(Field::new(name, DataType::Utf8));
 
         ChunkedArray {
             field,
-            chunks: vec![Arc::new(builder.finish())],
+            chunks: vec![builder.into_arc()],
             phantom: PhantomData,
             categorical_map: None,
             ..Default::default()
@@ -320,7 +305,7 @@ where
             Some(s) => acc + s.as_ref().len(),
             None => acc,
         });
-        let mut builder = Utf8ChunkedBuilder::new(name, values_size, opt_v.len());
+        let mut builder = Utf8ChunkedBuilder::new(name, opt_v.len(), values_size);
 
         opt_v.iter().for_each(|opt| match opt {
             Some(v) => builder.append_value(v.as_ref()),
@@ -354,105 +339,125 @@ pub trait ListBuilderTrait {
 
 pub struct ListPrimitiveChunkedBuilder<T>
 where
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
-    pub builder: LargeListBuilder<PrimitiveBuilder<T>>,
+    pub builder: LargePrimitiveBuilder<T::Native>,
     field: Field,
+    fast_explode: bool,
 }
 
 macro_rules! finish_list_builder {
     ($self:ident) => {{
-        let arr = Arc::new($self.builder.finish());
-        ListChunked {
+        let arr = $self.builder.as_arc();
+        let mut ca = ListChunked {
             field: Arc::new($self.field.clone()),
             chunks: vec![arr],
             phantom: PhantomData,
             categorical_map: None,
             ..Default::default()
+        };
+        if $self.fast_explode {
+            ca.set_fast_explode()
         }
+        ca
     }};
 }
 
 impl<T> ListPrimitiveChunkedBuilder<T>
 where
-    T: PolarsPrimitiveType,
+    T: PolarsNumericType,
 {
-    pub fn new(name: &str, values_builder: PrimitiveBuilder<T>, capacity: usize) -> Self {
-        let builder = LargeListBuilder::with_capacity(values_builder, capacity);
+    pub fn new(name: &str, capacity: usize, values_capacity: usize) -> Self {
+        let values = MutablePrimitiveArray::<T::Native>::with_capacity(values_capacity);
+        let builder = LargePrimitiveBuilder::<T::Native>::new_with_capacity(values, capacity);
         let field = Field::new(name, DataType::List(T::get_dtype().to_arrow()));
 
-        ListPrimitiveChunkedBuilder { builder, field }
+        Self {
+            builder,
+            field,
+            fast_explode: true,
+        }
     }
 
     pub fn append_slice(&mut self, opt_v: Option<&[T::Native]>) {
         match opt_v {
-            Some(v) => {
-                self.builder.values().append_slice(v).unwrap();
-                self.builder.append(true).expect("should not fail");
+            Some(items) => {
+                let values = self.builder.mut_values();
+                values.extend_from_slice(items);
+                self.builder.try_push_valid().unwrap();
+
+                if items.is_empty() {
+                    self.fast_explode = false;
+                }
             }
             None => {
-                self.builder.append(false).expect("should not fail");
+                self.builder.push_null();
             }
         }
     }
+    /// Appends from an iterator over values
+    #[inline]
+    pub fn append_iter_values<I: Iterator<Item = T::Native> + TrustedLen>(&mut self, iter: I) {
+        let values = self.builder.mut_values();
+
+        if iter.size_hint().0 == 0 {
+            self.fast_explode = false;
+        }
+        // Safety
+        // trusted len, trust the type system
+        unsafe { values.extend_trusted_len_values_unchecked(iter) };
+        self.builder.try_push_valid().unwrap();
+    }
 
     pub fn append_null(&mut self) {
-        self.builder.append(false).expect("should not fail");
+        self.builder.push_null();
     }
 }
 
 impl<T> ListBuilderTrait for ListPrimitiveChunkedBuilder<T>
 where
-    T: PolarsPrimitiveType,
-    T::Native: Num,
+    T: PolarsNumericType,
 {
     #[inline]
     fn append_opt_series(&mut self, opt_s: Option<&Series>) {
         match opt_s {
-            Some(s) => self.append_series(s),
+            Some(s) => {
+                self.append_series(s);
+            }
             None => {
-                self.builder.append(false).unwrap();
+                self.builder.push_null();
             }
         }
     }
 
     #[inline]
     fn append_null(&mut self) {
-        let builder = self.builder.values();
-        builder.append_null().unwrap();
-        self.builder.append(true).unwrap();
+        self.builder.push_null();
     }
 
     #[inline]
     fn append_series(&mut self, s: &Series) {
-        let builder = self.builder.values();
-        let arrays = s.chunks();
-        if s.null_count() == s.len() {
-            for _ in 0..s.len() {
-                builder.append_null().unwrap();
-            }
-        } else if s.is_empty() {
-            return;
-        } else {
-            for a in arrays {
-                let values = a.get_values::<T>();
-                // we would like to check if array has no null values.
-                // however at the time of writing there is a bug in append_slice, because it does not update
-                // the null bitmap
-                if s.null_count() == 0 {
-                    builder.append_slice(values).unwrap();
-                } else {
-                    values.iter().enumerate().for_each(|(idx, v)| {
-                        if a.is_valid(idx) {
-                            builder.append_value(*v).unwrap();
-                        } else {
-                            builder.append_null().unwrap();
-                        }
-                    });
-                }
-            }
+        if s.is_empty() {
+            self.fast_explode = false;
         }
-        self.builder.append(true).unwrap();
+        let arrays = s.chunks();
+        let values = self.builder.mut_values();
+
+        arrays.iter().for_each(|x| {
+            let arr = x
+                .as_any()
+                .downcast_ref::<PrimitiveArray<T::Native>>()
+                .unwrap();
+
+            if arr.null_count() == 0 {
+                values.extend_from_slice(arr.values().as_slice())
+            } else {
+                // Safety:
+                // Arrow arrays are trusted length iterators.
+                unsafe { values.extend_trusted_len_unchecked(arr.into_iter()) }
+            }
+        });
+        self.builder.try_push_valid().unwrap();
     }
 
     fn finish(&mut self) -> ListChunked {
@@ -460,17 +465,27 @@ where
     }
 }
 
+type LargePrimitiveBuilder<T> = MutableListArray<i64, MutablePrimitiveArray<T>>;
+type LargeListUtf8Builder = MutableListArray<i64, MutableUtf8Array<i64>>;
+type LargeListBooleanBuilder = MutableListArray<i64, MutableBooleanArray>;
+
 pub struct ListUtf8ChunkedBuilder {
-    builder: LargeListBuilder<LargeStringBuilder>,
+    builder: LargeListUtf8Builder,
     field: Field,
+    fast_explode: bool,
 }
 
 impl ListUtf8ChunkedBuilder {
-    pub fn new(name: &str, values_builder: LargeStringBuilder, capacity: usize) -> Self {
-        let builder = LargeListBuilder::with_capacity(values_builder, capacity);
+    pub fn new(name: &str, capacity: usize, values_capacity: usize) -> Self {
+        let values = MutableUtf8Array::<i64>::with_capacity(values_capacity);
+        let builder = LargeListUtf8Builder::new_with_capacity(values, capacity);
         let field = Field::new(name, DataType::List(ArrowDataType::LargeUtf8));
 
-        ListUtf8ChunkedBuilder { builder, field }
+        ListUtf8ChunkedBuilder {
+            builder,
+            field,
+            fast_explode: true,
+        }
     }
 }
 
@@ -479,29 +494,25 @@ impl ListBuilderTrait for ListUtf8ChunkedBuilder {
         match opt_s {
             Some(s) => self.append_series(s),
             None => {
-                self.builder.append(false).unwrap();
+                self.builder.push_null();
             }
         }
     }
 
     #[inline]
     fn append_null(&mut self) {
-        let builder = self.builder.values();
-        builder.append_null().unwrap();
-        self.builder.append(true).unwrap();
+        self.builder.push_null();
     }
 
     #[inline]
     fn append_series(&mut self, s: &Series) {
-        let ca = s.utf8().unwrap();
-        let value_builder = self.builder.values();
-        for s in ca {
-            match s {
-                Some(s) => value_builder.append_value(s).unwrap(),
-                None => value_builder.append_null().unwrap(),
-            };
+        if s.is_empty() {
+            self.fast_explode = false;
         }
-        self.builder.append(true).unwrap();
+        let ca = s.utf8().unwrap();
+        let value_builder = self.builder.mut_values();
+        value_builder.try_extend(ca).unwrap();
+        self.builder.try_push_valid().unwrap();
     }
 
     fn finish(&mut self) -> ListChunked {
@@ -510,16 +521,22 @@ impl ListBuilderTrait for ListUtf8ChunkedBuilder {
 }
 
 pub struct ListBooleanChunkedBuilder {
-    builder: LargeListBuilder<BooleanArrayBuilder>,
+    builder: LargeListBooleanBuilder,
     field: Field,
+    fast_explode: bool,
 }
 
 impl ListBooleanChunkedBuilder {
-    pub fn new(name: &str, values_builder: BooleanArrayBuilder, capacity: usize) -> Self {
-        let builder = LargeListBuilder::with_capacity(values_builder, capacity);
+    pub fn new(name: &str, capacity: usize, values_capacity: usize) -> Self {
+        let values = MutableBooleanArray::with_capacity(values_capacity);
+        let builder = LargeListBooleanBuilder::new_with_capacity(values, capacity);
         let field = Field::new(name, DataType::List(ArrowDataType::Boolean));
 
-        Self { builder, field }
+        Self {
+            builder,
+            field,
+            fast_explode: true,
+        }
     }
 }
 
@@ -528,29 +545,25 @@ impl ListBuilderTrait for ListBooleanChunkedBuilder {
         match opt_s {
             Some(s) => self.append_series(s),
             None => {
-                self.builder.append(false).unwrap();
+                self.builder.push_null();
             }
         }
     }
 
     #[inline]
     fn append_null(&mut self) {
-        let builder = self.builder.values();
-        builder.append_null();
-        self.builder.append(true).unwrap();
+        self.builder.push_null();
     }
 
     #[inline]
     fn append_series(&mut self, s: &Series) {
         let ca = s.bool().unwrap();
-        let value_builder = self.builder.values();
-        for s in ca {
-            match s {
-                Some(s) => value_builder.append_value(s),
-                None => value_builder.append_null(),
-            };
+        if ca.is_empty() {
+            self.fast_explode = false;
         }
-        self.builder.append(true).unwrap();
+        let value_builder = self.builder.mut_values();
+        value_builder.extend(ca);
+        self.builder.try_push_valid().unwrap();
     }
 
     fn finish(&mut self) -> ListChunked {
@@ -566,23 +579,20 @@ pub fn get_list_builder(
 ) -> Box<dyn ListBuilderTrait> {
     macro_rules! get_primitive_builder {
         ($type:ty) => {{
-            let values_builder = PrimitiveBuilder::<$type>::new(value_capacity);
-            let builder = ListPrimitiveChunkedBuilder::new(&name, values_builder, list_capacity);
+            let builder =
+                ListPrimitiveChunkedBuilder::<$type>::new(&name, list_capacity, value_capacity);
             Box::new(builder)
         }};
     }
     macro_rules! get_bool_builder {
         () => {{
-            let values_builder = BooleanArrayBuilder::new(value_capacity);
-            let builder = ListBooleanChunkedBuilder::new(&name, values_builder, list_capacity);
+            let builder = ListBooleanChunkedBuilder::new(&name, list_capacity, value_capacity);
             Box::new(builder)
         }};
     }
     macro_rules! get_utf8_builder {
         () => {{
-            let values_builder =
-                LargeStringBuilder::with_capacity(value_capacity * 5, value_capacity);
-            let builder = ListUtf8ChunkedBuilder::new(&name, values_builder, list_capacity);
+            let builder = ListUtf8ChunkedBuilder::new(&name, list_capacity, 5 * value_capacity);
             Box::new(builder)
         }};
     }
@@ -612,8 +622,7 @@ mod test {
 
     #[test]
     fn test_list_builder() {
-        let values_builder = PrimitiveBuilder::<Int32Type>::new(10);
-        let mut builder = ListPrimitiveChunkedBuilder::new("a", values_builder, 10);
+        let mut builder = ListPrimitiveChunkedBuilder::<Int32Type>::new("a", 10, 5);
 
         // create a series containing two chunks
         let mut s1 = Int32Chunked::new_from_slice("a", &[1, 2, 3]).into_series();
@@ -642,8 +651,7 @@ mod test {
 
     #[test]
     fn test_list_str_builder() {
-        let mut builder =
-            ListUtf8ChunkedBuilder::new("a", LargeStringBuilder::with_capacity(10, 10), 10);
+        let mut builder = ListUtf8ChunkedBuilder::new("a", 10, 10);
         builder.append_series(&Series::new("", &["foo", "bar"]));
         let ca = builder.finish();
         dbg!(ca);
