@@ -1,16 +1,8 @@
 //! The typed heart of every Series column.
-use crate::chunked_array::builder::get_bitmap;
 use crate::prelude::*;
-use arrow::{
-    array::{
-        ArrayRef, BooleanArray, Date64Array, Float32Array, Float64Array, Int16Array, Int32Array,
-        Int64Array, Int8Array, LargeStringArray, Time64NanosecondArray, UInt16Array, UInt32Array,
-        UInt64Array, UInt8Array,
-    },
-    buffer::Buffer,
-    datatypes::TimeUnit,
-};
+use arrow::{array::*, bitmap::Bitmap, datatypes::TimeUnit};
 use itertools::Itertools;
+use polars_arrow::prelude::ValueSize;
 use std::convert::TryFrom;
 use std::iter::{Copied, Map};
 use std::marker::PhantomData;
@@ -43,16 +35,12 @@ pub mod strings;
 pub mod temporal;
 mod trusted_len;
 pub mod upstream_traits;
-
-use arrow::array::{
-    Array, ArrayData, Date32Array, DurationMillisecondArray, DurationNanosecondArray,
-    LargeListArray, PrimitiveArray,
-};
+use arrow::array::Array;
+pub(crate) mod list;
+use polars_arrow::prelude::*;
 
 use crate::chunked_array::builder::categorical::RevMapping;
 use crate::utils::{slice_offsets, CustomIterTools};
-use arrow::util::bit_util::{get_bit, round_upto_power_of_2};
-use polars_arrow::array::ValueSize;
 use std::mem;
 use std::ops::{Deref, DerefMut};
 
@@ -153,8 +141,11 @@ pub struct ChunkedArray<T> {
     phantom: PhantomData<T>,
     /// maps categorical u32 indexes to String values
     pub(crate) categorical_map: Option<Arc<RevMapping>>,
-    // first bit: sorted
-    // second_bit: sorted reverse
+    /// first bit: sorted
+    /// second_bit: sorted reverse
+    /// third bit dtype list: fast_explode
+    ///     - unset: unknown or not all arrays have at least one value
+    ///     - set: all list arrays are filled (this allows for cheap explode)
     pub(crate) bit_settings: u8,
 }
 
@@ -176,12 +167,6 @@ impl<T> ChunkedArray<T> {
             self.bit_settings |= 1
         }
     }
-
-    /// Get Arrow ArrayData
-    pub fn array_data(&self) -> Vec<&ArrayData> {
-        self.chunks.iter().map(|arr| arr.data()).collect()
-    }
-
     /// Get a reference to the mapping of categorical types to the string values.
     pub fn get_categorical_map(&self) -> Option<&Arc<RevMapping>> {
         self.categorical_map.as_ref()
@@ -189,52 +174,36 @@ impl<T> ChunkedArray<T> {
 
     /// Get the index of the first non null value in this ChunkedArray.
     pub fn first_non_null(&self) -> Option<usize> {
-        if self.null_count() == self.len() {
-            None
-        } else if self.null_count() == 0 {
-            Some(0)
-        } else {
-            let mut offset = 0;
-            for (idx, (null_count, null_bit_buffer)) in self.null_bits().enumerate() {
-                if null_count == 0 {
-                    return Some(offset);
-                } else {
-                    let arr = &self.chunks[idx];
-                    let null_bit_buffer = null_bit_buffer.unwrap();
-                    let bit_end = arr.offset() + arr.len();
-
-                    let byte_start = std::cmp::min(round_upto_power_of_2(arr.offset(), 8), bit_end);
-                    let data = null_bit_buffer.as_slice();
-
-                    for i in arr.offset()..byte_start {
-                        if get_bit(data, i) {
-                            return Some(offset + i);
-                        }
+        let mut offset = 0;
+        for (_, null_bitmap) in self.null_bits() {
+            if let Some(null_bitmap) = null_bitmap {
+                for (idx, is_valid) in null_bitmap.iter().enumerate() {
+                    if is_valid {
+                        return Some(offset + idx);
                     }
-                    offset += arr.len()
                 }
+                offset += null_bitmap.len()
+            } else {
+                return Some(offset);
             }
-            None
         }
+        None
     }
 
-    /// Get the null count and the buffer of bits representing the validity values
-    pub fn null_bits(&self) -> impl Iterator<Item = (usize, Option<&Buffer>)> + '_ {
-        self.chunks.iter().map(|arr| get_bitmap(arr.as_ref()))
+    /// Get the buffer of bits representing null values
+    pub fn null_bits(&self) -> impl Iterator<Item = (usize, &Option<Bitmap>)> + '_ {
+        self.chunks
+            .iter()
+            .map(|arr| (arr.null_count(), arr.validity()))
     }
 
     /// Shrink the capacity of this array to fit it's length.
     pub fn shrink_to_fit(&mut self) {
-        if self.chunks.len() == 1 {
-            self.chunks = vec![polars_arrow::kernels::shrink_to_fit::shrink_to_fit(
-                &*self.chunks[0],
-            )]
-        }
-        self.chunks =
-            vec![
-                arrow::compute::concat(self.chunks.iter().map(|a| &**a).collect_vec().as_slice())
-                    .unwrap(),
-            ];
+        self.chunks = vec![arrow::compute::concat::concatenate(
+            self.chunks.iter().map(|a| &**a).collect_vec().as_slice(),
+        )
+        .unwrap()
+        .into()];
     }
 
     /// Unpack a Series to the same physical type.
@@ -314,6 +283,7 @@ impl<T> ChunkedArray<T> {
     }
 
     /// Count the null values.
+    #[inline]
     pub fn null_count(&self) -> usize {
         self.chunks.iter().map(|arr| arr.null_count()).sum()
     }
@@ -361,7 +331,7 @@ impl<T> ChunkedArray<T> {
             chunks,
             phantom: PhantomData,
             categorical_map: self.categorical_map.clone(),
-            ..Default::default()
+            bit_settings: self.bit_settings,
         }
     }
 
@@ -390,7 +360,7 @@ impl<T> ChunkedArray<T> {
                 take_len = remaining_length;
             }
 
-            new_chunks.push(chunk.slice(remaining_offset, take_len));
+            new_chunks.push(chunk.slice(remaining_offset, take_len).into());
             remaining_length -= take_len;
             remaining_offset = 0;
             if remaining_length == 0 {
@@ -408,12 +378,19 @@ impl<T> ChunkedArray<T> {
         let chunks = self
             .chunks
             .iter()
-            .map(|arr| Arc::new((&**arr).is_null_mask()) as ArrayRef)
+            .map(|arr| {
+                let bitmap = arr
+                    .validity()
+                    .as_ref()
+                    .map(|bitmap| !bitmap)
+                    .unwrap_or_else(|| Bitmap::new_zeroed(arr.len()));
+                Arc::new(BooleanArray::from_data_default(bitmap, None)) as ArrayRef
+            })
             .collect_vec();
         BooleanChunked::new_from_chunks("is_null", chunks)
     }
 
-    /// Get a mask of the null values.
+    /// Get a mask of the valid values.
     pub fn is_not_null(&self) -> BooleanChunked {
         if self.null_count() == 0 {
             return BooleanChunked::full("is_not_null", true, self.len());
@@ -421,7 +398,13 @@ impl<T> ChunkedArray<T> {
         let chunks = self
             .chunks
             .iter()
-            .map(|arr| Arc::new((&**arr).is_not_null_mask()) as ArrayRef)
+            .map(|arr| {
+                let bitmap = arr
+                    .validity()
+                    .clone()
+                    .unwrap_or_else(|| !(&Bitmap::new_zeroed(arr.len())));
+                Arc::new(BooleanArray::from_data_default(bitmap, None)) as ArrayRef
+            })
             .collect_vec();
         BooleanChunked::new_from_chunks("is_not_null", chunks)
     }
@@ -453,11 +436,16 @@ impl<T> ChunkedArray<T> {
     where
         Self: std::marker::Sized,
     {
-        if matches!(self.dtype(), DataType::Categorical) && !self.is_empty() {
-            assert!(Arc::ptr_eq(
-                self.categorical_map.as_ref().unwrap(),
-                other.categorical_map.as_ref().unwrap()
-            ));
+        if let (Some(rev_map_l), Some(rev_map_r)) = (
+            self.categorical_map.as_ref(),
+            other.categorical_map.as_ref(),
+        ) {
+            // first assertion checks if the global string cache is equal,
+            // the second checks if we append a slice from this array to self
+            if !rev_map_l.same_src(rev_map_r) && !Arc::ptr_eq(rev_map_l, rev_map_r) {
+                panic!("Appending categorical data can only be done if they are made under the same global string cache. \
+                Consider using a global string cache.")
+            }
         }
 
         // replace an empty array
@@ -504,7 +492,7 @@ where
             let mut offset = 0;
             let chunks = chunk_id
                 .map(|len| {
-                    let out = array.slice(offset, len);
+                    let out = array.slice(offset, len).into();
                     offset += len;
                     out
                 })
@@ -557,7 +545,7 @@ where
         macro_rules! downcast_and_pack {
             ($casttype:ident, $variant:ident) => {{
                 let arr = &*(arr as *const dyn Array as *const $casttype);
-                let v = arr.value_unchecked(idx);
+                let v = arr.value(idx);
                 AnyValue::$variant(v)
             }};
         }
@@ -581,22 +569,22 @@ where
             DataType::Int64 => downcast_and_pack!(Int64Array, Int64),
             DataType::Float32 => downcast_and_pack!(Float32Array, Float32),
             DataType::Float64 => downcast_and_pack!(Float64Array, Float64),
-            DataType::Date32 => downcast_and_pack!(Date32Array, Date32),
-            DataType::Date64 => downcast_and_pack!(Date64Array, Date64),
+            DataType::Date32 => downcast_and_pack!(Int32Array, Date32),
+            DataType::Date64 => downcast_and_pack!(Int64Array, Date64),
             DataType::Time64(TimeUnit::Nanosecond) => {
-                let v = downcast!(Time64NanosecondArray);
+                let v = downcast!(Int64Array);
                 AnyValue::Time64(v, TimeUnit::Nanosecond)
             }
             DataType::Duration(TimeUnit::Nanosecond) => {
-                let v = downcast!(DurationNanosecondArray);
+                let v = downcast!(Int64Array);
                 AnyValue::Duration(v, TimeUnit::Nanosecond)
             }
             DataType::Duration(TimeUnit::Millisecond) => {
-                let v = downcast!(DurationMillisecondArray);
+                let v = downcast!(Int64Array);
                 AnyValue::Duration(v, TimeUnit::Millisecond)
             }
             DataType::List(_) => {
-                let v = downcast!(LargeListArray);
+                let v: ArrayRef = downcast!(LargeListArray).into();
                 let s = Series::try_from(("", v));
                 AnyValue::List(s.unwrap())
             }
@@ -617,17 +605,17 @@ where
 {
     /// Create a new ChunkedArray by taking ownership of the AlignedVec. This operation is zero copy.
     pub fn new_from_aligned_vec(name: &str, v: AlignedVec<T::Native>) -> Self {
-        let arr = v.into_primitive_array::<T>(None);
-        Self::new_from_chunks(name, vec![Arc::new(arr)])
+        let arr = to_array::<T>(v, None);
+        Self::new_from_chunks(name, vec![arr])
     }
 
     /// Nullify values in slice with an existing null bitmap
     pub fn new_from_owned_with_null_bitmap(
         name: &str,
         values: AlignedVec<T::Native>,
-        buffer: Option<Buffer>,
+        buffer: Option<Bitmap>,
     ) -> Self {
-        let arr = Arc::new(values.into_primitive_array::<T>(buffer));
+        let arr = to_array::<T>(values, buffer);
         ChunkedArray {
             field: Arc::new(Field::new(name, T::get_dtype())),
             chunks: vec![arr],
@@ -684,7 +672,7 @@ where
     /// NOTE: null values should be taken into account by the user of these slices as they are handled
     /// separately
     pub fn data_views(&self) -> impl Iterator<Item = &[T::Native]> + DoubleEndedIterator {
-        self.downcast_iter().map(|arr| arr.values())
+        self.downcast_iter().map(|arr| arr.values().as_slice())
     }
 
     #[allow(clippy::wrong_self_convention)]
@@ -807,7 +795,7 @@ impl<T> Clone for ChunkedArray<T> {
             chunks: self.chunks.clone(),
             phantom: PhantomData,
             categorical_map: self.categorical_map.clone(),
-            ..Default::default()
+            bit_settings: self.bit_settings,
         }
     }
 }
@@ -875,8 +863,22 @@ impl ListChunked {
     }
 }
 
-impl<T: PolarsNumericType> From<PrimitiveArray<T>> for ChunkedArray<T> {
-    fn from(a: PrimitiveArray<T>) -> Self {
+pub(crate) fn to_primitive<T: PolarsPrimitiveType>(
+    values: AlignedVec<T::Native>,
+    validity: Option<Bitmap>,
+) -> PrimitiveArray<T::Native> {
+    PrimitiveArray::from_data(T::get_dtype().to_arrow(), values.into(), validity)
+}
+
+pub(crate) fn to_array<T: PolarsPrimitiveType>(
+    values: AlignedVec<T::Native>,
+    validity: Option<Bitmap>,
+) -> ArrayRef {
+    Arc::new(to_primitive::<T>(values, validity))
+}
+
+impl<T: PolarsNumericType> From<PrimitiveArray<T::Native>> for ChunkedArray<T> {
+    fn from(a: PrimitiveArray<T::Native>) -> Self {
         ChunkedArray::new_from_chunks("", vec![Arc::new(a)])
     }
 }
@@ -1073,6 +1075,7 @@ pub(crate) mod test {
     }
 
     #[test]
+    #[ignore]
     fn test_shrink_to_fit() {
         let mut builder = Utf8ChunkedBuilder::new("foo", 2048, 100 * 2048);
         builder.append_value("foo");
@@ -1080,15 +1083,14 @@ pub(crate) mod test {
         let before = arr
             .chunks()
             .iter()
-            .map(|arr| arr.get_buffer_memory_size())
+            .map(|arr| arrow::compute::aggregate::estimated_bytes_size(arr.as_ref()))
             .sum::<usize>();
         arr.shrink_to_fit();
         let after = arr
             .chunks()
             .iter()
-            .map(|arr| arr.get_buffer_memory_size())
+            .map(|arr| arrow::compute::aggregate::estimated_bytes_size(arr.as_ref()))
             .sum::<usize>();
-        dbg!(before, after);
         assert!(before > after);
     }
 }
