@@ -2,9 +2,7 @@
 use crate::chunked_array::object::ObjectArray;
 use crate::prelude::*;
 use crate::utils::index_to_chunked_index;
-use arrow::array::{
-    Array, ArrayRef, BooleanArray, LargeListArray, LargeStringArray, PrimitiveArray,
-};
+use arrow::array::*;
 use std::marker::PhantomData;
 
 pub struct Chunks<'a, T> {
@@ -42,13 +40,17 @@ impl<T> ChunkedArray<T>
 where
     T: PolarsNumericType,
 {
-    pub fn downcast_iter(&self) -> impl Iterator<Item = &PrimitiveArray<T>> + DoubleEndedIterator {
+    pub fn downcast_iter(
+        &self,
+    ) -> impl Iterator<Item = &PrimitiveArray<T::Native>> + DoubleEndedIterator {
         self.chunks.iter().map(|arr| {
+            // Safety:
+            // This should be the array type in PolarsNumericType
             let arr = &**arr;
-            unsafe { &*(arr as *const dyn Array as *const PrimitiveArray<T>) }
+            unsafe { &*(arr as *const dyn Array as *const PrimitiveArray<T::Native>) }
         })
     }
-    pub fn downcast_chunks(&self) -> Chunks<'_, PrimitiveArray<T>> {
+    pub fn downcast_chunks(&self) -> Chunks<'_, PrimitiveArray<T::Native>> {
         Chunks::new(&self.chunks)
     }
 
@@ -65,6 +67,8 @@ where
 impl BooleanChunked {
     pub fn downcast_iter(&self) -> impl Iterator<Item = &BooleanArray> + DoubleEndedIterator {
         self.chunks.iter().map(|arr| {
+            // Safety:
+            // This should be the array type in BooleanChunked
             let arr = &**arr;
             unsafe { &*(arr as *const dyn Array as *const BooleanArray) }
         })
@@ -83,13 +87,17 @@ impl BooleanChunked {
 }
 
 impl Utf8Chunked {
-    pub fn downcast_iter(&self) -> impl Iterator<Item = &LargeStringArray> + DoubleEndedIterator {
+    pub fn downcast_iter(&self) -> impl Iterator<Item = &Utf8Array<i64>> + DoubleEndedIterator {
+        // Safety:
+        // This is the array type that must be in a Utf8Chunked
         self.chunks.iter().map(|arr| {
+            // Safety:
+            // This should be the array type in Utf8Chunked
             let arr = &**arr;
-            unsafe { &*(arr as *const dyn Array as *const LargeStringArray) }
+            unsafe { &*(arr as *const dyn Array as *const Utf8Array<i64>) }
         })
     }
-    pub fn downcast_chunks(&self) -> Chunks<'_, LargeStringArray> {
+    pub fn downcast_chunks(&self) -> Chunks<'_, Utf8Array<i64>> {
         Chunks::new(&self.chunks)
     }
 
@@ -103,13 +111,15 @@ impl Utf8Chunked {
 }
 
 impl ListChunked {
-    pub fn downcast_iter(&self) -> impl Iterator<Item = &LargeListArray> + DoubleEndedIterator {
+    pub fn downcast_iter(&self) -> impl Iterator<Item = &ListArray<i64>> + DoubleEndedIterator {
+        // Safety:
+        // This is the array type that must be in a ListChunked
         self.chunks.iter().map(|arr| {
             let arr = &**arr;
-            unsafe { &*(arr as *const dyn Array as *const LargeListArray) }
+            unsafe { &*(arr as *const dyn Array as *const ListArray<i64>) }
         })
     }
-    pub fn downcast_chunks(&self) -> Chunks<'_, LargeListArray> {
+    pub fn downcast_chunks(&self) -> Chunks<'_, ListArray<i64>> {
         Chunks::new(&self.chunks)
     }
 
