@@ -1,6 +1,9 @@
+use crate::prelude::compare_inner::PartialOrdInner;
 use crate::prelude::*;
-use crate::utils::NoNull;
+use crate::utils::{CustomIterTools, NoNull};
 use itertools::Itertools;
+use polars_arrow::prelude::ValueSize;
+use polars_arrow::trusted_len::PushUnchecked;
 use rayon::prelude::*;
 use std::cmp::Ordering;
 use std::iter::FromIterator;
@@ -9,7 +12,7 @@ use std::ops::{Deref, DerefMut};
 /// Sort with null values, to reverse, swap the arguments.
 fn sort_with_nulls<T: PartialOrd>(a: &Option<T>, b: &Option<T>) -> Ordering {
     match (a, b) {
-        (Some(a), Some(b)) => a.partial_cmp(b).expect("could not compare"),
+        (Some(a), Some(b)) => order_default(a, b),
         (None, Some(_)) => Ordering::Less,
         (Some(_), None) => Ordering::Greater,
         (None, None) => Ordering::Equal,
@@ -104,53 +107,52 @@ macro_rules! argsort {
     ($self:expr, $reverse:expr) => {{
         let sort_parallel = sort_parallel($self);
 
-        let ca: NoNull<UInt32Chunked> = if $self.null_count() == 0 {
-            let mut count: u32 = 0;
-            let mut vals: Vec<_> = $self
-                .into_no_null_iter()
-                .map(|v| {
-                    let i = count;
-                    count += 1;
-                    (i, v)
-                })
-                .collect();
+        let mut vals = Vec::with_capacity($self.len());
+        let mut count: u32 = 0;
+        $self.downcast_iter().for_each(|arr| {
+            let iter = arr.iter().map(|v| {
+                let i = count;
+                count += 1;
+                (i, v)
+            });
+            vals.extend_trusted_len(iter);
+        });
 
-            argsort_branch(
-                vals.as_mut_slice(),
-                sort_parallel,
-                $reverse,
-                |(_, a), (_, b)| a.partial_cmp(b).unwrap(),
-                |(_, a), (_, b)| b.partial_cmp(a).unwrap(),
-            );
-
-            vals.into_iter().map(|(idx, _v)| idx).collect()
-        } else {
-            let mut count: u32 = 0;
-            let mut vals: Vec<_> = $self
-                .into_iter()
-                .map(|v| {
-                    let i = count;
-                    count += 1;
-                    (i, v)
-                })
-                .collect();
-
-            argsort_branch(
-                vals.as_mut_slice(),
-                sort_parallel,
-                $reverse,
-                |(_, a), (_, b)| order_default_null(a, b),
-                |(_, a), (_, b)| order_reverse_null(a, b),
-            );
-
-            vals.into_iter().map(|(idx, _v)| idx).collect()
-        };
+        argsort_branch(
+            vals.as_mut_slice(),
+            sort_parallel,
+            $reverse,
+            |(_, a), (_, b)| order_default_null(a, b),
+            |(_, a), (_, b)| order_reverse_null(a, b),
+        );
+        let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect_trusted();
         let mut ca = ca.into_inner();
         ca.rename($self.name());
         ca
     }};
 }
 
+fn memcpy_values<T>(ca: &ChunkedArray<T>) -> AlignedVec<T::Native>
+where
+    T: PolarsNumericType,
+{
+    let len = ca.len();
+    let mut vals = AlignedVec::with_capacity(len);
+    // Safety:
+    // only primitives so no drop calls when writing
+    unsafe { vals.set_len(len) }
+    let vals_slice = vals.as_mut_slice();
+
+    let mut offset = 0;
+    ca.downcast_iter().for_each(|arr| {
+        let values = arr.values();
+        let len = arr.len();
+        (vals_slice[offset..len]).copy_from_slice(values);
+        offset += len;
+    });
+    vals
+}
+
 impl<T> ChunkSort<T> for ChunkedArray<T>
 where
     T: PolarsNumericType,
@@ -159,45 +161,34 @@ where
     fn sort(&self, reverse: bool) -> ChunkedArray<T> {
         let sort_parallel = sort_parallel(self);
 
-        if let Ok(vals) = self.cont_slice() {
-            // Copy the values to a new aligned vec. This can be mutably sorted.
-            let n = self.len();
-            let vals_ptr = vals.as_ptr();
-            // allocate aligned
-            let mut new = AlignedVec::<T::Native>::with_capacity_aligned(n);
-            let new_ptr = new.as_mut_ptr();
-
-            // memcopy
-            unsafe { std::ptr::copy_nonoverlapping(vals_ptr, new_ptr, n) };
-            // set len to copied bytes
-            unsafe { new.set_len(n) };
+        if self.null_count() == 0 {
+            let mut vals = memcpy_values(self);
 
             sort_branch(
-                new.as_mut_slice(),
+                vals.as_mut_slice(),
                 sort_parallel,
                 reverse,
                 order_default,
                 order_reverse,
             );
 
-            return ChunkedArray::new_from_aligned_vec(self.name(), new);
-        }
-
-        if self.null_count() == 0 {
-            // rechunk and call again, then it will fall in the contiguous slice path.
-            let ca = self.rechunk();
-            ca.sort(reverse)
+            return ChunkedArray::new_from_aligned_vec(self.name(), vals);
         } else {
-            let mut v = Vec::from_iter(self);
+            let mut vals = Vec::with_capacity(self.len());
+            self.downcast_iter().for_each(|arr| {
+                let iter = arr.iter().map(|opt_v| opt_v.copied());
+                vals.extend_trusted_len(iter);
+            });
             sort_branch(
-                v.as_mut_slice(),
+                vals.as_mut_slice(),
                 sort_parallel,
                 reverse,
                 order_default_null,
                 order_reverse_null,
             );
-            let mut ca: Self = v.into_iter().collect();
+            let mut ca: Self = vals.into_iter().collect_trusted();
             ca.rename(self.name());
+            ca.set_sorted(reverse);
             ca
         }
     }
@@ -208,7 +199,54 @@ where
     }
 
     fn argsort(&self, reverse: bool) -> UInt32Chunked {
-        argsort!(self, reverse)
+        let sort_parallel = sort_parallel(self);
+
+        let ca: NoNull<UInt32Chunked> = if self.null_count() == 0 {
+            let mut vals = Vec::with_capacity(self.len());
+            let mut count: u32 = 0;
+            self.downcast_iter().for_each(|arr| {
+                let values = arr.values();
+                let iter = values.iter().map(|&v| {
+                    let i = count;
+                    count += 1;
+                    (i, v)
+                });
+                vals.extend_trusted_len(iter);
+            });
+
+            argsort_branch(
+                vals.as_mut_slice(),
+                sort_parallel,
+                reverse,
+                |(_, a), (_, b)| a.partial_cmp(b).unwrap(),
+                |(_, a), (_, b)| b.partial_cmp(a).unwrap(),
+            );
+
+            vals.into_iter().map(|(idx, _v)| idx).collect_trusted()
+        } else {
+            let mut vals = Vec::with_capacity(self.len());
+            let mut count: u32 = 0;
+            self.downcast_iter().for_each(|arr| {
+                let iter = arr.iter().map(|v| {
+                    let i = count;
+                    count += 1;
+                    (i, v)
+                });
+                vals.extend_trusted_len(iter);
+            });
+
+            argsort_branch(
+                vals.as_mut_slice(),
+                sort_parallel,
+                reverse,
+                |(_, a), (_, b)| order_default_null(a, b),
+                |(_, a), (_, b)| order_reverse_null(a, b),
+            );
+            vals.into_iter().map(|(idx, _v)| idx).collect_trusted()
+        };
+        let mut ca = ca.into_inner();
+        ca.rename(self.name());
+        ca
     }
 
     #[cfg(feature = "sort_multiple")]
@@ -232,6 +270,9 @@ where
         }
 
         assert_eq!(other.len(), reverse.len() - 1);
+
+        let compare_inner: Vec<_> = other.iter().map(|s| s.into_partial_ord_inner()).collect();
+
         let mut count: u32 = 0;
         let mut vals: Vec<_> = self
             .into_iter()
@@ -249,68 +290,7 @@ where
                 (_, Ordering::Equal) => {
                     let idx_a = tpl_a.0 as usize;
                     let idx_b = tpl_b.0 as usize;
-
-                    macro_rules! partial_ord_by_idx {
-                        ($ca: ident, $reverse: expr) => {{
-                            // Safety:
-                            // Indexes are in bounds, we asserted equal lengths above
-                            let a;
-                            let b;
-                            if $reverse {
-                                b = unsafe { $ca.get_unchecked(idx_a) };
-                                a = unsafe { $ca.get_unchecked(idx_b) };
-                            } else {
-                                a = unsafe { $ca.get_unchecked(idx_a) };
-                                b = unsafe { $ca.get_unchecked(idx_b) };
-                            }
-
-                            match (&a).partial_cmp(&b).unwrap() {
-                                // also equal, try next array
-                                Ordering::Equal => continue,
-                                // this array is not equal, return
-                                ord => return ord,
-                            }
-                        }};
-                    }
-
-                    // series should be matching type or utf8
-                    for (s, reverse) in other.iter().zip(&reverse[1..]) {
-                        match s.dtype() {
-                            DataType::Utf8 => {
-                                let ca = s.utf8().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Float32 => {
-                                let ca = s.f32().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Float64 => {
-                                let ca = s.f64().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Int64 => {
-                                let ca = s.i64().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Int32 => {
-                                let ca = s.i32().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::UInt32 => {
-                                let ca = s.u32().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::UInt64 => {
-                                let ca = s.u64().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            _ => {
-                                unreachable!()
-                            }
-                        }
-                    }
-                    // all arrays exhausted, ordering equal it is.
-                    Ordering::Equal
+                    ordering_other_columns(&compare_inner, &reverse[1..], idx_a, idx_b)
                 }
                 (true, Ordering::Less) => Ordering::Greater,
                 (true, Ordering::Greater) => Ordering::Less,
@@ -318,11 +298,32 @@ where
             },
         );
         let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect();
-
-        Ok(ca.into_inner())
+        let mut ca = ca.into_inner();
+        ca.set_sorted(reverse[0]);
+        Ok(ca)
     }
 }
 
+fn ordering_other_columns<'a>(
+    compare_inner: &'a [Box<dyn PartialOrdInner + 'a>],
+    reverse: &[bool],
+    idx_a: usize,
+    idx_b: usize,
+) -> Ordering {
+    for (cmp, reverse) in compare_inner.iter().zip(reverse) {
+        // Safety:
+        // indices are in bounds
+        let ordering = unsafe { cmp.cmp_element_unchecked(idx_a, idx_b) };
+        match (ordering, reverse) {
+            (Ordering::Equal, _) => continue,
+            (_, true) => return ordering.reverse(),
+            _ => return ordering,
+        }
+    }
+    // all arrays/columns exhausted, ordering equal it is.
+    Ordering::Equal
+}
+
 macro_rules! sort {
     ($self:ident, $reverse:ident) => {{
         if $reverse {
@@ -349,7 +350,9 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
         // We don't collect from an iterator because we know the total value size
         let mut builder = Utf8ChunkedBuilder::new(self.name(), self.len(), self.get_values_size());
         v.into_iter().for_each(|opt_v| builder.append_option(opt_v));
-        builder.finish()
+        let mut ca = builder.finish();
+        ca.set_sorted(reverse);
+        ca
     }
 
     fn sort_in_place(&mut self, reverse: bool) {
@@ -388,6 +391,7 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
                 (i, v)
             })
             .collect();
+        let compare_inner: Vec<_> = other.iter().map(|s| s.into_partial_ord_inner()).collect();
 
         vals.sort_by(
             |tpl_a, tpl_b| match (reverse[0], sort_with_nulls(&tpl_a.1, &tpl_b.1)) {
@@ -396,68 +400,7 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
                 (_, Ordering::Equal) => {
                     let idx_a = tpl_a.0 as usize;
                     let idx_b = tpl_b.0 as usize;
-
-                    macro_rules! partial_ord_by_idx {
-                        ($ca: ident, $reverse: expr) => {{
-                            // Safety:
-                            // Indexes are in bounds, we asserted equal lengths above
-                            let a;
-                            let b;
-                            if $reverse {
-                                b = unsafe { $ca.get_unchecked(idx_a) };
-                                a = unsafe { $ca.get_unchecked(idx_b) };
-                            } else {
-                                a = unsafe { $ca.get_unchecked(idx_a) };
-                                b = unsafe { $ca.get_unchecked(idx_b) };
-                            }
-
-                            match (&a).partial_cmp(&b).unwrap() {
-                                // also equal, try next array
-                                Ordering::Equal => continue,
-                                // this array is not equal, return
-                                ord => return ord,
-                            }
-                        }};
-                    }
-
-                    // series should be matching type or utf8
-                    for (s, reverse) in other.iter().zip(&reverse[1..]) {
-                        match s.dtype() {
-                            DataType::Utf8 => {
-                                let ca = s.utf8().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Float32 => {
-                                let ca = s.f32().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Float64 => {
-                                let ca = s.f64().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Int64 => {
-                                let ca = s.i64().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::Int32 => {
-                                let ca = s.i32().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::UInt32 => {
-                                let ca = s.u32().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            DataType::UInt64 => {
-                                let ca = s.u64().unwrap();
-                                partial_ord_by_idx!(ca, *reverse)
-                            }
-                            _ => {
-                                unreachable!()
-                            }
-                        }
-                    }
-                    // all arrays exhausted, ordering equal it is.
-                    Ordering::Equal
+                    ordering_other_columns(&compare_inner, &reverse[1..], idx_a, idx_b)
                 }
                 (true, Ordering::Less) => Ordering::Greater,
                 (true, Ordering::Greater) => Ordering::Less,
@@ -465,8 +408,9 @@ impl ChunkSort<Utf8Type> for Utf8Chunked {
             },
         );
         let ca: NoNull<UInt32Chunked> = vals.into_iter().map(|(idx, _v)| idx).collect();
-
-        Ok(ca.into_inner())
+        let mut ca = ca.into_inner();
+        ca.set_sorted(reverse[0]);
+        Ok(ca)
     }
 }
 
@@ -563,6 +507,7 @@ mod test {
 
     #[test]
     #[cfg(feature = "sort_multiple")]
+    #[cfg_attr(miri, ignore)]
     fn test_argsort_multiple() -> Result<()> {
         let a = Int32Chunked::new_from_slice("a", &[1, 2, 1, 1, 3, 4, 3, 3]);
         let b = Int64Chunked::new_from_slice("b", &[0, 1, 2, 3, 4, 5, 6, 1]);
