@@ -10,16 +10,8 @@
 use crate::chunked_array::object::PolarsObjectSafe;
 use crate::prelude::*;
 use ahash::RandomState;
-pub use arrow::datatypes::DataType as ArrowDataType;
-pub use arrow::datatypes::{
-    ArrowNumericType, ArrowPrimitiveType, BooleanType, Date32Type, Date64Type,
-    DurationMicrosecondType, DurationMillisecondType, DurationNanosecondType, DurationSecondType,
-    Field as ArrowField, Float32Type, Float64Type, Int16Type, Int32Type, Int64Type, Int8Type,
-    IntervalDayTimeType, IntervalUnit, IntervalYearMonthType, Schema as ArrowSchema,
-    Time32MillisecondType, Time32SecondType, Time64MicrosecondType, Time64NanosecondType, TimeUnit,
-    TimestampMicrosecondType, TimestampMillisecondType, TimestampNanosecondType,
-    TimestampSecondType, UInt16Type, UInt32Type, UInt64Type, UInt8Type,
-};
+pub use arrow::datatypes::{DataType as ArrowDataType, TimeUnit};
+use arrow::types::NativeType;
 #[cfg(feature = "serde")]
 use serde::{Deserialize, Serialize};
 use std::cmp::Ordering;
@@ -33,11 +25,15 @@ pub struct ListType {}
 pub struct CategoricalType {}
 
 pub trait PolarsDataType: Send + Sync {
-    fn get_dtype() -> DataType;
+    fn get_dtype() -> DataType
+    where
+        Self: Sized;
 }
 
 macro_rules! impl_polars_datatype {
-    ($ca:ident, $variant:ident) => {
+    ($ca:ident, $variant:ident, $physical:ty) => {
+        pub struct $ca {}
+
         impl PolarsDataType for $ca {
             fn get_dtype() -> DataType {
                 DataType::$variant
@@ -46,19 +42,20 @@ macro_rules! impl_polars_datatype {
     };
 }
 
-impl_polars_datatype!(UInt8Type, UInt8);
-impl_polars_datatype!(UInt16Type, UInt16);
-impl_polars_datatype!(UInt32Type, UInt32);
-impl_polars_datatype!(UInt64Type, UInt64);
-impl_polars_datatype!(Int8Type, Int8);
-impl_polars_datatype!(Int16Type, Int16);
-impl_polars_datatype!(Int32Type, Int32);
-impl_polars_datatype!(Int64Type, Int64);
-impl_polars_datatype!(Float32Type, Float32);
-impl_polars_datatype!(Float64Type, Float64);
-impl_polars_datatype!(BooleanType, Boolean);
-impl_polars_datatype!(Date32Type, Date32);
-impl_polars_datatype!(Date64Type, Date64);
+impl_polars_datatype!(UInt8Type, UInt8, u8);
+impl_polars_datatype!(UInt16Type, UInt16, u16);
+impl_polars_datatype!(UInt32Type, UInt32, u32);
+impl_polars_datatype!(UInt64Type, UInt64, u64);
+impl_polars_datatype!(Int8Type, Int8, i8);
+impl_polars_datatype!(Int16Type, Int16, i16);
+impl_polars_datatype!(Int32Type, Int32, i32);
+impl_polars_datatype!(Int64Type, Int64, i64);
+impl_polars_datatype!(Float32Type, Float32, f32);
+impl_polars_datatype!(Float64Type, Float64, f64);
+impl_polars_datatype!(Date32Type, Date32, i32);
+impl_polars_datatype!(Date64Type, Date64, i64);
+
+pub struct Time64NanosecondType {}
 
 impl PolarsDataType for Time64NanosecondType {
     fn get_dtype() -> DataType {
@@ -66,12 +63,16 @@ impl PolarsDataType for Time64NanosecondType {
     }
 }
 
+pub struct DurationNanosecondType {}
+
 impl PolarsDataType for DurationNanosecondType {
     fn get_dtype() -> DataType {
         DataType::Duration(TimeUnit::Nanosecond)
     }
 }
 
+pub struct DurationMillisecondType {}
+
 impl PolarsDataType for DurationMillisecondType {
     fn get_dtype() -> DataType {
         DataType::Duration(TimeUnit::Millisecond)
@@ -84,6 +85,14 @@ impl PolarsDataType for Utf8Type {
     }
 }
 
+pub struct BooleanType {}
+
+impl PolarsDataType for BooleanType {
+    fn get_dtype() -> DataType {
+        DataType::Boolean
+    }
+}
+
 impl PolarsDataType for ListType {
     fn get_dtype() -> DataType {
         // null as we cannot no anything without self.
@@ -114,7 +123,7 @@ impl<T: PolarsObject> PolarsDataType for ObjectType<T> {
 /// Any type that is not nested
 pub trait PolarsSingleType: PolarsDataType {}
 
-impl<T> PolarsSingleType for T where T: ArrowPrimitiveType + PolarsDataType {}
+impl<T> PolarsSingleType for T where T: NativeType + PolarsDataType {}
 
 impl PolarsSingleType for Utf8Type {}
 
@@ -138,40 +147,77 @@ pub type DurationMillisecondChunked = ChunkedArray<DurationMillisecondType>;
 pub type Time64NanosecondChunked = ChunkedArray<Time64NanosecondType>;
 pub type CategoricalChunked = ChunkedArray<CategoricalType>;
 
-pub trait PolarsPrimitiveType: ArrowPrimitiveType + Send + Sync + PolarsDataType {}
-// impl PolarsPrimitiveType for BooleanType {}
-impl PolarsPrimitiveType for UInt8Type {}
-impl PolarsPrimitiveType for UInt16Type {}
-impl PolarsPrimitiveType for UInt32Type {}
-impl PolarsPrimitiveType for UInt64Type {}
-impl PolarsPrimitiveType for Int8Type {}
-impl PolarsPrimitiveType for Int16Type {}
-impl PolarsPrimitiveType for Int32Type {}
-impl PolarsPrimitiveType for Int64Type {}
-impl PolarsPrimitiveType for Float32Type {}
-impl PolarsPrimitiveType for Float64Type {}
-impl PolarsPrimitiveType for Date32Type {}
-impl PolarsPrimitiveType for Date64Type {}
-impl PolarsPrimitiveType for Time64NanosecondType {}
-impl PolarsPrimitiveType for DurationNanosecondType {}
-impl PolarsPrimitiveType for DurationMillisecondType {}
+pub trait PolarsPrimitiveType: Send + Sync + PolarsDataType + 'static {
+    type Native: NativeType;
+}
+impl PolarsPrimitiveType for UInt8Type {
+    type Native = u8;
+}
+impl PolarsPrimitiveType for UInt16Type {
+    type Native = u16;
+}
+impl PolarsPrimitiveType for UInt32Type {
+    type Native = u32;
+}
+impl PolarsPrimitiveType for UInt64Type {
+    type Native = u64;
+}
+impl PolarsPrimitiveType for Int8Type {
+    type Native = i8;
+}
+impl PolarsPrimitiveType for Int16Type {
+    type Native = i16;
+}
+impl PolarsPrimitiveType for Int32Type {
+    type Native = i32;
+}
+impl PolarsPrimitiveType for Int64Type {
+    type Native = i64;
+}
+impl PolarsPrimitiveType for Float32Type {
+    type Native = f32;
+}
+impl PolarsPrimitiveType for Float64Type {
+    type Native = f64;
+}
+impl PolarsPrimitiveType for Date32Type {
+    type Native = i32;
+}
+impl PolarsPrimitiveType for Date64Type {
+    type Native = i64;
+}
+impl PolarsPrimitiveType for Time64NanosecondType {
+    type Native = i64;
+}
+impl PolarsPrimitiveType for DurationNanosecondType {
+    type Native = i64;
+}
+impl PolarsPrimitiveType for DurationMillisecondType {
+    type Native = i64;
+}
 
-pub trait PolarsNumericType: PolarsPrimitiveType + ArrowNumericType {}
-impl PolarsNumericType for UInt8Type {}
-impl PolarsNumericType for UInt16Type {}
-impl PolarsNumericType for UInt32Type {}
-impl PolarsNumericType for UInt64Type {}
-impl PolarsNumericType for Int8Type {}
-impl PolarsNumericType for Int16Type {}
-impl PolarsNumericType for Int32Type {}
-impl PolarsNumericType for Int64Type {}
-impl PolarsNumericType for Float32Type {}
-impl PolarsNumericType for Float64Type {}
-impl PolarsNumericType for Date32Type {}
-impl PolarsNumericType for Date64Type {}
-impl PolarsNumericType for Time64NanosecondType {}
-impl PolarsNumericType for DurationNanosecondType {}
-impl PolarsNumericType for DurationMillisecondType {}
+macro_rules! impl_polars_numeric {
+    ($ca:ident, $physical:ty) => {
+        impl PolarsNumericType for $ca {}
+    };
+}
+
+pub trait PolarsNumericType: PolarsPrimitiveType {}
+impl_polars_numeric!(UInt8Type, u8);
+impl_polars_numeric!(UInt16Type, u16);
+impl_polars_numeric!(UInt32Type, u32);
+impl_polars_numeric!(UInt64Type, u64);
+impl_polars_numeric!(Int8Type, i8);
+impl_polars_numeric!(Int16Type, i16);
+impl_polars_numeric!(Int32Type, i32);
+impl_polars_numeric!(Int64Type, i64);
+impl_polars_numeric!(Float32Type, f32);
+impl_polars_numeric!(Float64Type, f64);
+impl_polars_numeric!(Date32Type, i32);
+impl_polars_numeric!(Date64Type, i64);
+impl_polars_numeric!(Time64NanosecondType, i64);
+impl_polars_numeric!(DurationNanosecondType, i64);
+impl_polars_numeric!(DurationMillisecondType, i64);
 
 pub trait PolarsIntegerType: PolarsNumericType {}
 impl PolarsIntegerType for UInt8Type {}
@@ -192,6 +238,31 @@ pub trait PolarsFloatType: PolarsNumericType {}
 impl PolarsFloatType for Float32Type {}
 impl PolarsFloatType for Float64Type {}
 
+pub trait ToPolarsType {
+    fn to_polars_type() -> &'static dyn PolarsDataType;
+}
+
+macro_rules! to_polars_type {
+    ($native:ty, $polars:ident) => {
+        impl ToPolarsType for $native {
+            fn to_polars_type() -> &'static dyn PolarsDataType {
+                &$polars {}
+            }
+        }
+    };
+}
+
+to_polars_type!(i8, Int8Type);
+to_polars_type!(i16, Int16Type);
+to_polars_type!(i32, Int32Type);
+to_polars_type!(i64, Int64Type);
+to_polars_type!(u8, UInt8Type);
+to_polars_type!(u16, UInt16Type);
+to_polars_type!(u32, UInt32Type);
+to_polars_type!(u64, UInt64Type);
+to_polars_type!(f32, Float32Type);
+to_polars_type!(f64, Float64Type);
+
 #[derive(Debug, Clone)]
 pub enum AnyValue<'a> {
     Null,
@@ -461,13 +532,13 @@ impl DataType {
             Utf8 => ArrowDataType::LargeUtf8,
             Date32 => ArrowDataType::Date32,
             Date64 => ArrowDataType::Date64,
-            Time64(tu) => ArrowDataType::Time64(tu.clone()),
+            Time64(tu) => ArrowDataType::Time64(*tu),
             List(dt) => ArrowDataType::LargeList(Box::new(arrow::datatypes::Field::new(
                 "",
                 dt.clone(),
                 true,
             ))),
-            Duration(tu) => ArrowDataType::Duration(tu.clone()),
+            Duration(tu) => ArrowDataType::Duration(*tu),
             Null => ArrowDataType::Null,
             #[cfg(feature = "object")]
             Object(_) => panic!("cannot convert object to arrow"),
