@@ -12,9 +12,7 @@ use std::{
 const LIMIT: usize = 25;
 
 #[cfg(feature = "pretty_fmt")]
-use comfy_table::modifiers::UTF8_ROUND_CORNERS;
-#[cfg(feature = "pretty_fmt")]
-use comfy_table::presets::UTF8_FULL;
+use comfy_table::presets::{ASCII_FULL, UTF8_FULL};
 #[cfg(feature = "pretty_fmt")]
 use comfy_table::*;
 
@@ -75,14 +73,14 @@ use temporal::*;
 
 macro_rules! format_array {
     ($limit:expr, $f:ident, $a:expr, $dtype:expr, $name:expr, $array_type:expr) => {{
-        write![
+        write!(
             $f,
             "shape: ({},)\n{}: '{}' [{}]\n[\n",
             $a.len(),
             $array_type,
             $name,
             $dtype
-        ]?;
+        )?;
         let truncate = matches!($a.dtype(), DataType::Utf8);
         let limit = std::cmp::min($limit, $a.len());
 
@@ -123,7 +121,7 @@ macro_rules! format_array {
             }
         }
 
-        write![$f, "]"]
+        write!($f, "]")
     }};
 }
 
@@ -137,21 +135,21 @@ fn format_object_array(
 ) -> fmt::Result {
     match object.dtype() {
         DataType::Object(inner_type) => {
-            write![
+            write!(
                 f,
                 "shape: ({},)\n{}: '{}' [o][{}]\n[\n",
                 object.len(),
                 array_type,
                 name,
                 inner_type
-            ]?;
+            )?;
 
             for i in 0..limit {
                 let v = object.str_value(i);
                 writeln!(f, "\t{}", v)?;
             }
 
-            write![f, "]"]
+            write!(f, "]")
         }
         _ => unreachable!(),
     }
@@ -211,12 +209,12 @@ where
 
         let taker = self.take_rand();
         let inner_type = T::type_name();
-        write![
+        write!(
             f,
             "ChunkedArray: '{}' [o][{}]\n[\n",
             self.name(),
             inner_type
-        ]?;
+        )?;
 
         if limit < self.len() {
             for i in 0..limit / 2 {
@@ -350,7 +348,7 @@ impl Debug for Series {
                 self.name(),
                 "Series"
             ),
-            _ => unimplemented!(),
+            dt => panic!("{:?} not impl", dt),
         }
     }
 }
@@ -439,10 +437,15 @@ impl Display for DataFrame {
         #[cfg(feature = "pretty_fmt")]
         {
             let mut table = Table::new();
+            let preset = if std::env::var("POLARS_FMT_NO_UTF8").is_ok() {
+                ASCII_FULL
+            } else {
+                UTF8_FULL
+            };
+
             table
-                .load_preset(UTF8_FULL)
+                .load_preset(preset)
                 .set_content_arrangement(ContentArrangement::Dynamic)
-                .apply_modifier(UTF8_ROUND_CORNERS)
                 .set_table_width(
                     std::env::var("POLARS_TABLE_WIDTH")
                         .map(|s| {
@@ -672,12 +675,10 @@ impl<T> FmtList for ObjectChunked<T> {
 ))]
 mod test {
     use crate::prelude::*;
-    use polars_arrow::prelude::PrimitiveArrayBuilder;
 
     #[test]
     fn test_fmt_list() {
-        let values_builder = PrimitiveArrayBuilder::<UInt32Type>::new(10);
-        let mut builder = ListPrimitiveChunkedBuilder::new("a", values_builder, 10);
+        let mut builder = ListPrimitiveChunkedBuilder::<Int32Type>::new("a", 10, 10);
         builder.append_slice(Some(&[1, 2, 3]));
         builder.append_slice(None);
         let list = builder.finish().into_series();
