@@ -5,16 +5,16 @@ use rayon::prelude::*;
 use std::collections::HashSet;
 use std::hash::Hash;
 
+use arrow::types::{simd::Simd, NativeType};
+
 use crate::chunked_array::kernels::take_agg::{
     take_agg_no_null_primitive_iter_unchecked, take_agg_primitive_iter_unchecked,
     take_agg_primitive_iter_unchecked_count_nulls,
 };
 use crate::prelude::*;
 use crate::utils::NoNull;
-use arrow::array::{Array, ArrayData, ArrayRef, LargeListArray, PrimitiveBuilder};
 use arrow::buffer::MutableBuffer;
-use polars_arrow::builder::{BooleanArrayBuilder, LargeStringBuilder};
-use std::convert::TryFrom;
+use polars_arrow::trusted_len::PushUnchecked;
 
 pub(crate) trait NumericAggSync {
     fn agg_mean(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
@@ -73,11 +73,20 @@ impl<T> NumericAggSync for ObjectChunked<T> {}
 impl<T> NumericAggSync for ChunkedArray<T>
 where
     T: PolarsNumericType + Sync,
-    T::Native: std::ops::Add<Output = T::Native> + Num + NumCast + Bounded,
+    T::Native: NativeType + PartialOrd + Num + NumCast + Zero + Simd + Bounded,
+    <T::Native as Simd>::Simd: std::ops::Add<Output = <T::Native as Simd>::Simd>
+        + arrow::compute::aggregate::Sum<T::Native>
+        + arrow::compute::aggregate::SimdOrd<T::Native>,
     ChunkedArray<T>: IntoSeries,
 {
     fn agg_mean(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<Float64Type, _>(groups, |(first, idx)| {
+            // this can fail due to a bug in lazy code.
+            // here users can create filters in aggregations
+            // and thereby creating shorter columns than the original group tuples.
+            // the group tuples are modified, but if that's done incorrect there can be out of bounds
+            // access
+            debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
                 None
             } else if idx.len() == 1 {
@@ -95,7 +104,7 @@ where
                     .to_f64()
                     .map(|sum| sum / idx.len() as f64),
                     (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked_count_nulls(
+                        take_agg_primitive_iter_unchecked_count_nulls::<T, _, _>(
                             self.downcast_iter().next().unwrap(),
                             idx.iter().map(|i| *i as usize),
                             |a, b| a + b,
@@ -120,6 +129,7 @@ where
 
     fn agg_min(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<T, _>(groups, |(first, idx)| {
+            debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
                 None
             } else if idx.len() == 1 {
@@ -135,7 +145,7 @@ where
                         )
                     }),
                     (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked(
+                        take_agg_primitive_iter_unchecked::<T, _, _>(
                             self.downcast_iter().next().unwrap(),
                             idx.iter().map(|i| *i as usize),
                             |a, b| if a < b { a } else { b },
@@ -154,6 +164,7 @@ where
 
     fn agg_max(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<T, _>(groups, |(first, idx)| {
+            debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
                 None
             } else if idx.len() == 1 {
@@ -169,7 +180,7 @@ where
                         )
                     }),
                     (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked(
+                        take_agg_primitive_iter_unchecked::<T, _, _>(
                             self.downcast_iter().next().unwrap(),
                             idx.iter().map(|i| *i as usize),
                             |a, b| if a > b { a } else { b },
@@ -188,6 +199,7 @@ where
 
     fn agg_sum(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<T, _>(groups, |(first, idx)| {
+            debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
                 None
             } else if idx.len() == 1 {
@@ -203,7 +215,7 @@ where
                         )
                     }),
                     (_, 1) => unsafe {
-                        take_agg_primitive_iter_unchecked(
+                        take_agg_primitive_iter_unchecked::<T, _, _>(
                             self.downcast_iter().next().unwrap(),
                             idx.iter().map(|i| *i as usize),
                             |a, b| a + b,
@@ -221,6 +233,7 @@ where
     }
     fn agg_var(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<Float64Type, _>(groups, |(_first, idx)| {
+            debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
                 return None;
             }
@@ -234,6 +247,7 @@ where
     }
     fn agg_std(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<Float64Type, _>(groups, |(_first, idx)| {
+            debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
                 return None;
             }
@@ -248,6 +262,7 @@ where
     #[cfg(feature = "lazy")]
     fn agg_valid_count(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
         agg_helper::<UInt32Type, _>(groups, |(_first, idx)| {
+            debug_assert!(idx.len() <= self.len());
             if idx.is_empty() {
                 None
             } else if self.null_count() == 0 {
@@ -269,6 +284,7 @@ macro_rules! impl_agg_first {
         let mut ca = $groups
             .iter()
             .map(|(first, idx)| {
+                debug_assert!(idx.len() <= $self.len());
                 if idx.is_empty() {
                     return None;
                 }
@@ -349,6 +365,7 @@ macro_rules! impl_agg_last {
         let mut ca = $groups
             .iter()
             .map(|(_first, idx)| {
+                debug_assert!(idx.len() <= $self.len());
                 if idx.is_empty() {
                     return None;
                 }
@@ -428,21 +445,23 @@ macro_rules! impl_agg_n_unique {
         $groups
             .into_par_iter()
             .map(|(_first, idx)| {
+                debug_assert!(idx.len() <= $self.len());
                 if idx.is_empty() {
                     return 0;
                 }
+                let taker = $self.take_rand();
 
                 if $self.null_count() == 0 {
                     let mut set = HashSet::with_hasher(RandomState::new());
                     for i in idx {
-                        let v = unsafe { $self.get_unchecked(*i as usize) };
+                        let v = unsafe { taker.get_unchecked(*i as usize) };
                         set.insert(v);
                     }
                     set.len() as u32
                 } else {
                     let mut set = HashSet::with_hasher(RandomState::new());
                     for i in idx {
-                        let opt_v = $self.get(*i as usize);
+                        let opt_v = taker.get(*i as usize);
                         set.insert(opt_v);
                     }
                     set.len() as u32
@@ -503,43 +522,43 @@ pub trait AggList {
 impl<T> AggList for ChunkedArray<T>
 where
     T: PolarsNumericType,
-    T::Native: Num,
     ChunkedArray<T>: IntoSeries,
 {
     fn agg_list(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        let mut can_fast_explode = true;
         let arr = match self.cont_slice() {
             Ok(values) => {
-                let mut offsets =
-                    MutableBuffer::new((groups.len() + 1) * std::mem::size_of::<i64>());
+                let mut offsets = MutableBuffer::<i64>::with_capacity(groups.len() + 1);
                 let mut length_so_far = 0i64;
                 offsets.push(length_so_far);
-                let mut av = AlignedVec::with_capacity_aligned(self.len());
 
+                let mut list_values = MutableBuffer::<T::Native>::with_capacity(self.len());
                 groups.iter().for_each(|(_, idx)| {
-                    length_so_far += idx.len() as i64;
+                    let idx_len = idx.len();
+                    if idx_len == 0 {
+                        can_fast_explode = false;
+                    }
+
+                    length_so_far += idx_len as i64;
                     // Safety:
                     // group tuples are in bounds
                     unsafe {
-                        av.extend(idx.iter().map(|idx| *values.get_unchecked(*idx as usize)));
+                        list_values.extend_from_trusted_len_iter(
+                            idx.iter().map(|idx| *values.get_unchecked(*idx as usize)),
+                        );
+                        // Safety:
+                        // we know that offsets has allocated enough slots
+                        offsets.push_unchecked(length_so_far);
                     }
-                    offsets.push(length_so_far);
                 });
-                let values = av.into_primitive_array::<T>(None);
-
-                let field = Box::new(arrow::datatypes::Field::new("item", T::DATA_TYPE, true));
-                let data_type = ArrowDataType::LargeList(field);
-                let data = ArrayData::builder(data_type)
-                    .len(groups.len())
-                    .add_buffer(offsets.into())
-                    .add_child_data(values.data().clone())
-                    .build();
-                LargeListArray::from(data)
+                let array =
+                    PrimitiveArray::from_data(T::get_dtype().to_arrow(), list_values.into(), None);
+                let data_type = ListArray::<i64>::default_datatype(T::get_dtype().to_arrow());
+                ListArray::<i64>::from_data(data_type, offsets.into(), Arc::new(array), None)
             }
             _ => {
-                let values_cap = self.len();
-                let values_builder = PrimitiveBuilder::<T>::new(values_cap);
                 let mut builder =
-                    ListPrimitiveChunkedBuilder::new(self.name(), values_builder, groups.len());
+                    ListPrimitiveChunkedBuilder::<T>::new(self.name(), groups.len(), self.len());
                 for (_first, idx) in groups {
                     let s = unsafe {
                         self.take_unchecked(idx.iter().map(|i| *i as usize).into())
@@ -550,15 +569,17 @@ where
                 return Some(builder.finish().into_series());
             }
         };
-        Series::try_from((self.name(), Arc::new(arr) as ArrayRef)).ok()
+        let mut ca = ListChunked::new_from_chunks(self.name(), vec![Arc::new(arr)]);
+        if can_fast_explode {
+            ca.set_fast_explode()
+        }
+        Some(ca.into())
     }
 }
 
 impl AggList for BooleanChunked {
     fn agg_list(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        let values_cap = self.len();
-        let values_builder = BooleanArrayBuilder::new(values_cap);
-        let mut builder = ListBooleanChunkedBuilder::new(self.name(), values_builder, groups.len());
+        let mut builder = ListBooleanChunkedBuilder::new(self.name(), groups.len(), self.len());
         for (_first, idx) in groups {
             let s = unsafe {
                 self.take_unchecked(idx.iter().map(|i| *i as usize).into())
@@ -572,9 +593,7 @@ impl AggList for BooleanChunked {
 
 impl AggList for Utf8Chunked {
     fn agg_list(&self, groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-        let values_cap = self.len();
-        let values_builder = LargeStringBuilder::with_capacity(values_cap * 5, values_cap);
-        let mut builder = ListUtf8ChunkedBuilder::new(self.name(), values_builder, groups.len());
+        let mut builder = ListUtf8ChunkedBuilder::new(self.name(), groups.len(), self.len());
         for (_first, idx) in groups {
             let s = unsafe {
                 self.take_unchecked(idx.iter().map(|i| *i as usize).into())
@@ -603,7 +622,10 @@ pub(crate) trait AggQuantile {
 impl<T> AggQuantile for ChunkedArray<T>
 where
     T: PolarsNumericType + Sync,
-    T::Native: PartialOrd + Num + NumCast + Zero,
+    T::Native: PartialOrd + Num + NumCast + Zero + Simd,
+    <T::Native as Simd>::Simd: std::ops::Add<Output = <T::Native as Simd>::Simd>
+        + arrow::compute::aggregate::Sum<T::Native>
+        + arrow::compute::aggregate::SimdOrd<T::Native>,
     ChunkedArray<T>: IntoSeries,
 {
     fn agg_quantile(&self, groups: &[(u32, Vec<u32>)], quantile: f64) -> Option<Series> {
@@ -613,11 +635,7 @@ where
             }
 
             let group_vals = unsafe { self.take_unchecked(idx.iter().map(|i| *i as usize).into()) };
-            let sorted_idx_ca = group_vals.argsort(false);
-            let sorted_idx = sorted_idx_ca.downcast_iter().next().unwrap().values();
-            let quant_idx = (quantile * (sorted_idx.len() - 1) as f64) as usize;
-            let value_idx = sorted_idx[quant_idx];
-            group_vals.get(value_idx as usize)
+            group_vals.quantile(quantile).unwrap()
         })
     }
 
