@@ -1,11 +1,12 @@
 mod multiple_keys;
+use polars_arrow::utils::CustomIterTools;
 
 use crate::frame::hash_join::multiple_keys::{
     inner_join_multiple_keys, left_join_multiple_keys, outer_join_multiple_keys,
 };
 use crate::frame::select::Selection;
 use crate::prelude::*;
-use crate::utils::{set_partition_size, split_ca, NoNull};
+use crate::utils::{set_partition_size, split_ca};
 use crate::vector_hasher::{
     create_hash_and_keys_threaded_vectorized, prepare_hashed_relation_threaded, this_partition,
     AsU64, StrHash,
@@ -373,8 +374,8 @@ fn hash_join_tuples_outer<T, I, J>(
     swap: bool,
 ) -> Vec<(Option<u32>, Option<u32>)>
 where
-    I: Iterator<Item = T> + Send,
-    J: Iterator<Item = T> + Send,
+    I: Iterator<Item = T> + Send + TrustedLen,
+    J: Iterator<Item = T> + Send + TrustedLen,
     T: Hash + Eq + Copy + Sync + Send,
 {
     // This function is partially multi-threaded.
@@ -491,7 +492,7 @@ impl HashJoin<CategoricalType> for CategoricalChunked {
 fn num_group_join_inner<T>(left: &ChunkedArray<T>, right: &ChunkedArray<T>) -> Vec<(u32, u32)>
 where
     T: PolarsIntegerType,
-    T::Native: Hash + Eq + Send + AsU64,
+    T::Native: Hash + Eq + Send + AsU64 + Copy,
     Option<T::Native>: AsU64,
 {
     let n_threads = POOL.current_num_threads();
@@ -531,7 +532,7 @@ where
                 .iter()
                 .map(|ca| {
                     ca.downcast_iter()
-                        .map(|v| v.into_iter().map(|v| v.as_u64()))
+                        .map(|v| v.into_iter().map(|v| v.copied().as_u64()))
                         .flatten()
                         .collect::<Vec<_>>()
                 })
@@ -541,7 +542,7 @@ where
                 .iter()
                 .map(|ca| {
                     ca.downcast_iter()
-                        .map(|v| v.into_iter().map(|v| v.as_u64()))
+                        .map(|v| v.into_iter().map(|v| v.copied().as_u64()))
                         .flatten()
                         .collect::<Vec<_>>()
                 })
@@ -594,11 +595,11 @@ where
         (0, 0, _, _) => {
             let keys_a = splitted_a
                 .iter()
-                .map(|ca| ca.into_no_null_iter().collect::<Vec<_>>())
+                .map(|ca| ca.into_no_null_iter().collect_trusted::<Vec<_>>())
                 .collect::<Vec<_>>();
             let keys_b = splitted_b
                 .iter()
-                .map(|ca| ca.into_no_null_iter().collect::<Vec<_>>())
+                .map(|ca| ca.into_no_null_iter().collect_trusted::<Vec<_>>())
                 .collect::<Vec<_>>();
             hash_join_tuples_left(keys_a, keys_b)
         }
@@ -607,9 +608,10 @@ where
                 .iter()
                 .map(|ca| {
                     ca.downcast_iter()
-                        .map(|v| v.into_iter().map(|v| v.as_u64()))
+                        .map(|v| v.into_iter().map(|v| v.copied().as_u64()))
                         .flatten()
-                        .collect::<Vec<_>>()
+                        .trust_my_length(ca.len())
+                        .collect_trusted::<Vec<_>>()
                 })
                 .collect::<Vec<_>>();
 
@@ -617,9 +619,10 @@ where
                 .iter()
                 .map(|ca| {
                     ca.downcast_iter()
-                        .map(|v| v.into_iter().map(|v| v.as_u64()))
+                        .map(|v| v.into_iter().map(|v| v.copied().as_u64()))
                         .flatten()
-                        .collect::<Vec<_>>()
+                        .trust_my_length(ca.len())
+                        .collect_trusted::<Vec<_>>()
                 })
                 .collect::<Vec<_>>();
             hash_join_tuples_left(keys_a, keys_b)
@@ -627,11 +630,19 @@ where
         _ => {
             let keys_a = splitted_a
                 .iter()
-                .map(|ca| ca.into_iter().map(|v| v.as_u64()).collect::<Vec<_>>())
+                .map(|ca| {
+                    ca.into_iter()
+                        .map(|v| v.as_u64())
+                        .collect_trusted::<Vec<_>>()
+                })
                 .collect::<Vec<_>>();
             let keys_b = splitted_b
                 .iter()
-                .map(|ca| ca.into_iter().map(|v| v.as_u64()).collect::<Vec<_>>())
+                .map(|ca| {
+                    ca.into_iter()
+                        .map(|v| v.as_u64())
+                        .collect_trusted::<Vec<_>>()
+                })
                 .collect::<Vec<_>>();
             hash_join_tuples_left(keys_a, keys_b)
         }
@@ -857,8 +868,14 @@ impl HashJoin<Utf8Type> for Utf8Chunked {
                 hash_join_tuples_outer(iters_a, iters_b, swap)
             }
             _ => {
-                let iters_a = splitted_a.iter().map(|ca| ca.into_iter()).collect_vec();
-                let iters_b = splitted_b.iter().map(|ca| ca.into_iter()).collect_vec();
+                let iters_a = splitted_a
+                    .iter()
+                    .map(|ca| ca.into_iter())
+                    .collect::<Vec<_>>();
+                let iters_b = splitted_b
+                    .iter()
+                    .map(|ca| ca.into_iter())
+                    .collect::<Vec<_>>();
                 hash_join_tuples_outer(iters_a, iters_b, swap)
             }
         }
@@ -902,8 +919,7 @@ where
                     }
                 }
             })
-            .collect::<NoNull<ChunkedArray<T>>>()
-            .into_inner()
+            .collect::<ChunkedArray<T>>()
             .into_series()
     }
 }
@@ -1001,14 +1017,28 @@ impl DataFrame {
             return self.cross_join(other);
         }
 
-        let selected_left = self.select_series(left_on)?;
-        let selected_right = other.select_series(right_on)?;
-        assert_eq!(selected_right.len(), selected_left.len());
+        #[allow(unused_mut)]
+        let mut selected_left = self.select_series(left_on)?;
+        #[allow(unused_mut)]
+        let mut selected_right = other.select_series(right_on)?;
+        if selected_right.len() != selected_left.len() {
+            return Err(PolarsError::ValueError(
+                "the number of columns given as join key should be equal".into(),
+            ));
+        }
+        if selected_left
+            .iter()
+            .zip(&selected_right)
+            .any(|(l, r)| l.dtype() != r.dtype())
+        {
+            return Err(PolarsError::ValueError("the dtype of the join keys don't match. first cast your columns to the correct dtype".into()));
+        }
 
         for (l, r) in selected_left.iter().zip(&selected_right) {
             check_categorical_src(l, r)?
         }
 
+        // Single keys
         if selected_left.len() == 1 {
             return match how {
                 JoinType::Inner => {
@@ -1042,14 +1072,33 @@ impl DataFrame {
             new.unwrap()
         }
 
+        // hack for a macro
         impl DataFrame {
             fn len(&self) -> usize {
                 self.height()
             }
         }
 
-        // This is still single threaded and can create very large keys that are inserted in the
-        // hashmap. TODO: implement same hashing technique as in grouping.
+        // check if we have floats, in that case we have to coerce them
+        if selected_left
+            .iter()
+            .any(|s| matches!(s.dtype(), DataType::Float32 | DataType::Float64))
+        {
+            #[cfg(feature = "dtype-u64")]
+            for selected in &mut [&mut selected_left, &mut selected_right] {
+                selected.iter_mut().for_each(|s| match s.dtype().clone() {
+                    DataType::Float64 => *s = s.bit_repr_large().into_series(),
+                    DataType::Float32 => *s = s.bit_repr_small().into_series(),
+                    _ => {}
+                });
+            }
+            #[cfg(not(feature = "dtype-u64"))]
+            {
+                panic!("activate dtype-u64 feature to be able to join on floats")
+            }
+        }
+
+        // multiple keys
         match how {
             JoinType::Inner => {
                 let left = DataFrame::new_no_checks(selected_left);
@@ -1551,4 +1600,75 @@ mod test {
         right.inner_join(&left, "key", "key").unwrap();
         right.outer_join(&left, "key", "key").unwrap();
     }
+
+    #[test]
+    #[cfg_attr(miri, ignore)]
+    fn unit_df_join() -> Result<()> {
+        let df1 = df![
+            "a" => [1],
+            "b" => [2]
+        ]?;
+
+        let df2 = df![
+            "a" => [1, 2, 3, 4],
+            "b" => [Some(1), None, Some(3), Some(4)]
+        ]?;
+
+        let out = df1.left_join(&df2, "a", "a")?;
+        let expected = df![
+            "a" => [1],
+            "b" => [2],
+            "b_right" => [1]
+        ]?;
+        assert!(out.frame_equal(&expected));
+        Ok(())
+    }
+
+    #[test]
+    #[cfg_attr(miri, ignore)]
+    fn test_join_err() -> Result<()> {
+        let df1 = df![
+            "a" => [1, 2],
+            "b" => ["foo", "bar"]
+        ]?;
+
+        let df2 = df![
+            "a" => [1, 2, 3, 4],
+            "b" => [true, true, true, false]
+        ]?;
+
+        // dtypes don't match, error
+        assert!(df1
+            .join(&df2, vec!["a", "b"], vec!["a", "b"], JoinType::Left)
+            .is_err());
+        // length of join keys don't match error
+        assert!(df1
+            .join(&df2, vec!["a"], vec!["a", "b"], JoinType::Left)
+            .is_err());
+        Ok(())
+    }
+
+    #[test]
+    #[cfg_attr(miri, ignore)]
+    #[cfg(feature = "dtype-u64")]
+    fn test_join_floats() -> Result<()> {
+        let df_a = df! {
+            "a" => &[1.0, 2.0, 1.0, 1.0],
+            "b" => &["a", "b", "c", "c"],
+            "c" => &[0.0, 1.0, 2.0, 3.0]
+        }?;
+
+        let df_b = df! {
+            "foo" => &[1.0, 2.0, 1.0],
+            "bar" => &[1.0, 1.0, 1.0],
+            "ham" => &["let", "var", "const"]
+        }?;
+
+        let out = df_a.join(&df_b, vec!["a", "c"], vec!["foo", "bar"], JoinType::Left)?;
+        assert_eq!(
+            Vec::from(out.column("ham")?.utf8()?),
+            &[None, Some("var"), None, None]
+        );
+        Ok(())
+    }
 }
