@@ -10,6 +10,9 @@ mod fmt;
 pub mod frame;
 pub mod functions;
 pub mod prelude;
+#[cfg(feature = "serde")]
+#[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
+pub mod serde;
 pub mod series;
 pub mod testing;
 pub(crate) mod vector_hasher;
@@ -23,7 +26,11 @@ use std::sync::{Mutex, MutexGuard};
 // this is re-exported in utils for polars child crates
 lazy_static! {
     pub static ref POOL: ThreadPool = ThreadPoolBuilder::new()
-        .num_threads(num_cpus::get())
+        .num_threads(
+            std::env::var("POLARS_MAX_THREADS")
+                .map(|s| s.parse::<usize>().expect("integer"))
+                .unwrap_or_else(|_| num_cpus::get())
+        )
         .build()
         .expect("could not spawn threads");
 }
