@@ -17,8 +17,6 @@ use crate::frame::groupby::pivot::*;
 use crate::frame::groupby::*;
 use crate::prelude::*;
 use ahash::RandomState;
-use arrow::array::{ArrayData, ArrayRef};
-use arrow::buffer::Buffer;
 #[cfg(feature = "object")]
 use std::any::Any;
 use std::borrow::Cow;
@@ -107,6 +105,84 @@ macro_rules! impl_dyn_series {
         }
 
         impl private::PrivateSeries for SeriesWrap<$ca> {
+            fn explode_by_offsets(&self, offsets: &[i64]) -> Series {
+                physical_dispatch!(self, explode_by_offsets, offsets)
+            }
+
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_mean(
+                &self,
+                _window_size: u32,
+                _weight: Option<&[f64]>,
+                _ignore_null: bool,
+                _min_periods: u32,
+            ) -> Result<Series> {
+                Err(PolarsError::ComputeError(
+                    "cannot compute rolling mean of dates".into(),
+                ))
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_sum(
+                &self,
+                _window_size: u32,
+                _weight: Option<&[f64]>,
+                _ignore_null: bool,
+                _min_periods: u32,
+            ) -> Result<Series> {
+                Err(PolarsError::ComputeError(
+                    "cannot compute rolling sum of dates".into(),
+                ))
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_min(
+                &self,
+                window_size: u32,
+                weight: Option<&[f64]>,
+                ignore_null: bool,
+                min_periods: u32,
+            ) -> Result<Series> {
+                try_physical_dispatch!(
+                    self,
+                    rolling_min,
+                    window_size,
+                    weight,
+                    ignore_null,
+                    min_periods
+                )
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_max(
+                &self,
+                window_size: u32,
+                weight: Option<&[f64]>,
+                ignore_null: bool,
+                min_periods: u32,
+            ) -> Result<Series> {
+                try_physical_dispatch!(
+                    self,
+                    rolling_max,
+                    window_size,
+                    weight,
+                    ignore_null,
+                    min_periods
+                )
+            }
+
+            #[cfg(feature = "cum_agg")]
+            fn _cummax(&self, reverse: bool) -> Series {
+                physical_dispatch!(self, cummax, reverse)
+            }
+
+            #[cfg(feature = "cum_agg")]
+            fn _cummin(&self, reverse: bool) -> Series {
+                physical_dispatch!(self, cummin, reverse)
+            }
+
+            #[cfg(feature = "cum_agg")]
+            fn _cumsum(&self, _reverse: bool) -> Series {
+                panic!("cannot sum dates")
+            }
+
             #[cfg(feature = "asof_join")]
             fn join_asof(&self, other: &Series) -> Result<Vec<Option<u32>>> {
                 cast_and_apply!(self, join_asof, other)
@@ -247,7 +323,7 @@ macro_rules! impl_dyn_series {
                         let rhs = rhs.cast_with_dtype(&DataType::Int64).unwrap();
                         Ok(lhs.subtract(&rhs)?.into_series())
                     }
-                    (dtl, dtr) => Err(PolarsError::Other(
+                    (dtl, dtr) => Err(PolarsError::ComputeError(
                         format!(
                             "cannot do subtraction on these date types: {:?}, {:?}",
                             dtl, dtr
@@ -257,18 +333,22 @@ macro_rules! impl_dyn_series {
                 }
             }
             fn add_to(&self, _rhs: &Series) -> Result<Series> {
-                Err(PolarsError::Other("cannot do addition on dates".into()))
+                Err(PolarsError::ComputeError(
+                    "cannot do addition on dates".into(),
+                ))
             }
             fn multiply(&self, _rhs: &Series) -> Result<Series> {
-                Err(PolarsError::Other(
+                Err(PolarsError::ComputeError(
                     "cannot do multiplication on dates".into(),
                 ))
             }
             fn divide(&self, _rhs: &Series) -> Result<Series> {
-                Err(PolarsError::Other("cannot do division on dates".into()))
+                Err(PolarsError::ComputeError(
+                    "cannot do division on dates".into(),
+                ))
             }
             fn remainder(&self, _rhs: &Series) -> Result<Series> {
-                Err(PolarsError::Other(
+                Err(PolarsError::ComputeError(
                     "cannot do remainder operation on dates".into(),
                 ))
             }
@@ -292,26 +372,15 @@ macro_rules! impl_dyn_series {
         }
 
         impl SeriesTrait for SeriesWrap<$ca> {
-            fn cum_max(&self, reverse: bool) -> Series {
-                physical_dispatch!(self, cum_max, reverse)
-            }
-
-            fn cum_min(&self, reverse: bool) -> Series {
-                physical_dispatch!(self, cum_min, reverse)
-            }
-
-            fn cum_sum(&self, _reverse: bool) -> Series {
-                panic!("cannot sum dates")
+            #[cfg(feature = "interpolate")]
+            fn interpolate(&self) -> Series {
+                physical_dispatch!(self, interpolate,)
             }
 
             fn rename(&mut self, name: &str) {
                 self.0.rename(name);
             }
 
-            fn array_data(&self) -> Vec<&ArrayData> {
-                self.0.array_data()
-            }
-
             fn chunk_lengths(&self) -> ChunkIdIter {
                 self.0.chunk_id()
             }
@@ -444,19 +513,19 @@ macro_rules! impl_dyn_series {
                 try_physical_dispatch!(self, filter, filter)
             }
 
-            fn take(&self, indices: &UInt32Chunked) -> Series {
-                physical_dispatch!(self, take, indices)
+            fn take(&self, indices: &UInt32Chunked) -> Result<Series> {
+                try_physical_dispatch!(self, take, indices)
             }
 
-            fn take_iter(&self, iter: &mut dyn Iterator<Item = usize>) -> Series {
-                physical_dispatch!(self, take_iter, iter)
+            fn take_iter(&self, iter: &mut dyn TakeIterator) -> Result<Series> {
+                try_physical_dispatch!(self, take_iter, iter)
             }
 
             fn take_every(&self, n: usize) -> Series {
                 physical_dispatch!(self, take_every, n)
             }
 
-            unsafe fn take_iter_unchecked(&self, iter: &mut dyn Iterator<Item = usize>) -> Series {
+            unsafe fn take_iter_unchecked(&self, iter: &mut dyn TakeIterator) -> Series {
                 physical_dispatch!(self, take_iter_unchecked, iter)
             }
 
@@ -464,15 +533,13 @@ macro_rules! impl_dyn_series {
                 try_physical_dispatch!(self, take_unchecked, idx)
             }
 
-            unsafe fn take_opt_iter_unchecked(
-                &self,
-                iter: &mut dyn Iterator<Item = Option<usize>>,
-            ) -> Series {
+            unsafe fn take_opt_iter_unchecked(&self, iter: &mut dyn TakeIteratorNulls) -> Series {
                 physical_dispatch!(self, take_opt_iter_unchecked, iter)
             }
 
-            fn take_opt_iter(&self, iter: &mut dyn Iterator<Item = Option<usize>>) -> Series {
-                physical_dispatch!(self, take_opt_iter, iter)
+            #[cfg(feature = "take_opt_iter")]
+            fn take_opt_iter(&self, iter: &mut dyn TakeIteratorNulls) -> Result<Series> {
+                try_physical_dispatch!(self, take_opt_iter, iter)
             }
 
             fn len(&self) -> usize {
@@ -573,10 +640,6 @@ macro_rules! impl_dyn_series {
                 cast_and_apply!(self, is_duplicated,)
             }
 
-            fn null_bits(&self) -> Vec<(usize, Option<&Buffer>)> {
-                self.0.null_bits().collect()
-            }
-
             fn reverse(&self) -> Series {
                 physical_dispatch!(self, reverse,)
             }
@@ -589,12 +652,15 @@ macro_rules! impl_dyn_series {
                 physical_dispatch!(self, shift, periods)
             }
 
-            fn fill_none(&self, strategy: FillNoneStrategy) -> Result<Series> {
-                try_physical_dispatch!(self, fill_none, strategy)
+            fn fill_null(&self, strategy: FillNullStrategy) -> Result<Series> {
+                try_physical_dispatch!(self, fill_null, strategy)
             }
 
             fn sum_as_series(&self) -> Series {
-                panic!("cannot compute sum of dates")
+                Int32Chunked::full_null(self.name(), 1)
+                    .cast_with_dtype(self.dtype())
+                    .unwrap()
+                    .into()
             }
             fn max_as_series(&self) -> Series {
                 physical_dispatch!(self, max_as_series,)
@@ -603,13 +669,22 @@ macro_rules! impl_dyn_series {
                 physical_dispatch!(self, min_as_series,)
             }
             fn mean_as_series(&self) -> Series {
-                panic!("cannot compute mean of dates")
+                Int32Chunked::full_null(self.name(), 1)
+                    .cast_with_dtype(self.dtype())
+                    .unwrap()
+                    .into()
             }
             fn median_as_series(&self) -> Series {
-                panic!("cannot compute median of dates")
+                Int32Chunked::full_null(self.name(), 1)
+                    .cast_with_dtype(self.dtype())
+                    .unwrap()
+                    .into()
             }
             fn var_as_series(&self) -> Series {
-                panic!("cannot compute variance of dates")
+                Int32Chunked::full_null(self.name(), 1)
+                    .cast_with_dtype(self.dtype())
+                    .unwrap()
+                    .into()
             }
             fn std_as_series(&self) -> Series {
                 physical_dispatch!(self, std_as_series,)
@@ -617,65 +692,6 @@ macro_rules! impl_dyn_series {
             fn quantile_as_series(&self, quantile: f64) -> Result<Series> {
                 try_physical_dispatch!(self, quantile_as_series, quantile)
             }
-            fn rolling_mean(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                try_physical_dispatch!(
-                    self,
-                    rolling_mean,
-                    window_size,
-                    weight,
-                    ignore_null,
-                    min_periods
-                )
-            }
-            fn rolling_sum(
-                &self,
-                _window_size: u32,
-                _weight: Option<&[f64]>,
-                _ignore_null: bool,
-                _min_periods: u32,
-            ) -> Result<Series> {
-                Err(PolarsError::Other(
-                    "cannot compute rolling sum of dates".into(),
-                ))
-            }
-            fn rolling_min(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                try_physical_dispatch!(
-                    self,
-                    rolling_min,
-                    window_size,
-                    weight,
-                    ignore_null,
-                    min_periods
-                )
-            }
-            fn rolling_max(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                try_physical_dispatch!(
-                    self,
-                    rolling_max,
-                    window_size,
-                    weight,
-                    ignore_null,
-                    min_periods
-                )
-            }
 
             fn fmt_list(&self) -> String {
                 FmtList::fmt_list(&self.0)
@@ -698,7 +714,9 @@ macro_rules! impl_dyn_series {
             }
 
             fn pow(&self, _exponent: f64) -> Result<Series> {
-                Err(PolarsError::Other("cannot compute power of dates".into()))
+                Err(PolarsError::ComputeError(
+                    "cannot compute power of dates".into(),
+                ))
             }
 
             fn peak_max(&self) -> BooleanChunked {
