@@ -19,6 +19,7 @@ use crate::chunked_array::{
     ops::{
         aggregate::{ChunkAggSeries, VarAggSeries},
         compare_inner::{IntoPartialEqInner, IntoPartialOrdInner, PartialEqInner, PartialOrdInner},
+        explode::ExplodeByOffsets,
     },
     AsSinglePtr, ChunkIdIter,
 };
@@ -33,8 +34,7 @@ use crate::prelude::*;
 #[cfg(feature = "checked_arithmetic")]
 use crate::series::arithmetic::checked::NumOpsDispatchChecked;
 use ahash::RandomState;
-use arrow::array::{ArrayData, ArrayRef};
-use arrow::buffer::Buffer;
+use arrow::array::ArrayRef;
 use std::borrow::Cow;
 use std::ops::Deref;
 
@@ -64,6 +64,70 @@ macro_rules! impl_dyn_series {
         }
 
         impl private::PrivateSeries for SeriesWrap<$ca> {
+            fn explode_by_offsets(&self, offsets: &[i64]) -> Series {
+                self.0.explode_by_offsets(offsets)
+            }
+
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_mean(
+                &self,
+                window_size: u32,
+                weight: Option<&[f64]>,
+                ignore_null: bool,
+                min_periods: u32,
+            ) -> Result<Series> {
+                ChunkWindow::rolling_mean(&self.0, window_size, weight, ignore_null, min_periods)
+                    .map(|ca| ca.into_series())
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_sum(
+                &self,
+                window_size: u32,
+                weight: Option<&[f64]>,
+                ignore_null: bool,
+                min_periods: u32,
+            ) -> Result<Series> {
+                ChunkWindow::rolling_sum(&self.0, window_size, weight, ignore_null, min_periods)
+                    .map(|ca| ca.into_series())
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_min(
+                &self,
+                window_size: u32,
+                weight: Option<&[f64]>,
+                ignore_null: bool,
+                min_periods: u32,
+            ) -> Result<Series> {
+                ChunkWindow::rolling_min(&self.0, window_size, weight, ignore_null, min_periods)
+                    .map(|ca| ca.into_series())
+            }
+            #[cfg(feature = "rolling_window")]
+            fn _rolling_max(
+                &self,
+                window_size: u32,
+                weight: Option<&[f64]>,
+                ignore_null: bool,
+                min_periods: u32,
+            ) -> Result<Series> {
+                ChunkWindow::rolling_max(&self.0, window_size, weight, ignore_null, min_periods)
+                    .map(|ca| ca.into_series())
+            }
+
+            #[cfg(feature = "cum_agg")]
+            fn _cummax(&self, reverse: bool) -> Series {
+                self.0.cummax(reverse).into_series()
+            }
+
+            #[cfg(feature = "cum_agg")]
+            fn _cummin(&self, reverse: bool) -> Series {
+                self.0.cummin(reverse).into_series()
+            }
+
+            #[cfg(feature = "cum_agg")]
+            fn _cumsum(&self, reverse: bool) -> Series {
+                self.0.cumsum(reverse).into_series()
+            }
+
             #[cfg(feature = "asof_join")]
             fn join_asof(&self, other: &Series) -> Result<Vec<Option<u32>>> {
                 self.0.join_asof(other.as_ref().as_ref())
@@ -221,26 +285,24 @@ macro_rules! impl_dyn_series {
         }
 
         impl SeriesTrait for SeriesWrap<$ca> {
-            fn cum_max(&self, reverse: bool) -> Series {
-                self.0.cum_max(reverse).into_series()
+            #[cfg(feature = "rolling_window")]
+            fn rolling_apply(
+                &self,
+                _window_size: usize,
+                _f: &dyn Fn(&Series) -> Series,
+            ) -> Result<Series> {
+                ChunkRollApply::rolling_apply(&self.0, _window_size, _f).map(|ca| ca.into_series())
             }
 
-            fn cum_min(&self, reverse: bool) -> Series {
-                self.0.cum_min(reverse).into_series()
-            }
-
-            fn cum_sum(&self, reverse: bool) -> Series {
-                self.0.cum_sum(reverse).into_series()
+            #[cfg(feature = "interpolate")]
+            fn interpolate(&self) -> Series {
+                self.0.interpolate().into_series()
             }
 
             fn rename(&mut self, name: &str) {
                 self.0.rename(name);
             }
 
-            fn array_data(&self) -> Vec<&ArrayData> {
-                self.0.array_data()
-            }
-
             fn chunk_lengths(&self) -> ChunkIdIter {
                 self.0.chunk_id()
             }
@@ -583,24 +645,24 @@ macro_rules! impl_dyn_series {
                 self.0.median()
             }
 
-            fn take(&self, indices: &UInt32Chunked) -> Series {
+            fn take(&self, indices: &UInt32Chunked) -> Result<Series> {
                 let indices = if indices.chunks.len() > 1 {
                     Cow::Owned(indices.rechunk())
                 } else {
                     Cow::Borrowed(indices)
                 };
-                ChunkTake::take(&self.0, (&*indices).into()).into_series()
+                Ok(ChunkTake::take(&self.0, (&*indices).into())?.into_series())
             }
 
-            fn take_iter(&self, iter: &mut dyn Iterator<Item = usize>) -> Series {
-                ChunkTake::take(&self.0, iter.into()).into_series()
+            fn take_iter(&self, iter: &mut dyn TakeIterator) -> Result<Series> {
+                Ok(ChunkTake::take(&self.0, iter.into())?.into_series())
             }
 
             fn take_every(&self, n: usize) -> Series {
                 self.0.take_every(n).into_series()
             }
 
-            unsafe fn take_iter_unchecked(&self, iter: &mut dyn Iterator<Item = usize>) -> Series {
+            unsafe fn take_iter_unchecked(&self, iter: &mut dyn TakeIterator) -> Series {
                 ChunkTake::take_unchecked(&self.0, iter.into()).into_series()
             }
 
@@ -613,15 +675,13 @@ macro_rules! impl_dyn_series {
                 Ok(ChunkTake::take_unchecked(&self.0, (&*idx).into()).into_series())
             }
 
-            unsafe fn take_opt_iter_unchecked(
-                &self,
-                iter: &mut dyn Iterator<Item = Option<usize>>,
-            ) -> Series {
-                ChunkTake::take_unchecked(&self.0, SeriesWrap(iter).into()).into_series()
+            unsafe fn take_opt_iter_unchecked(&self, iter: &mut dyn TakeIteratorNulls) -> Series {
+                ChunkTake::take_unchecked(&self.0, iter.into()).into_series()
             }
 
-            fn take_opt_iter(&self, iter: &mut dyn Iterator<Item = Option<usize>>) -> Series {
-                ChunkTake::take(&self.0, SeriesWrap(iter).into()).into_series()
+            #[cfg(feature = "take_opt_iter")]
+            fn take_opt_iter(&self, iter: &mut dyn TakeIteratorNulls) -> Result<Series> {
+                Ok(ChunkTake::take(&self.0, iter.into())?.into_series())
             }
 
             fn len(&self) -> usize {
@@ -722,10 +782,6 @@ macro_rules! impl_dyn_series {
                 ChunkUnique::is_duplicated(&self.0)
             }
 
-            fn null_bits(&self) -> Vec<(usize, Option<&Buffer>)> {
-                self.0.null_bits().collect()
-            }
-
             fn reverse(&self) -> Series {
                 ChunkReverse::reverse(&self.0).into_series()
             }
@@ -738,8 +794,8 @@ macro_rules! impl_dyn_series {
                 ChunkShift::shift(&self.0, periods).into_series()
             }
 
-            fn fill_none(&self, strategy: FillNoneStrategy) -> Result<Series> {
-                ChunkFillNone::fill_none(&self.0, strategy).map(|ca| ca.into_series())
+            fn fill_null(&self, strategy: FillNullStrategy) -> Result<Series> {
+                ChunkFillNull::fill_null(&self.0, strategy).map(|ca| ca.into_series())
             }
 
             fn sum_as_series(&self) -> Series {
@@ -766,46 +822,6 @@ macro_rules! impl_dyn_series {
             fn quantile_as_series(&self, quantile: f64) -> Result<Series> {
                 ChunkAggSeries::quantile_as_series(&self.0, quantile)
             }
-            fn rolling_mean(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_mean(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
-            }
-            fn rolling_sum(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_sum(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
-            }
-            fn rolling_min(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_min(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
-            }
-            fn rolling_max(
-                &self,
-                window_size: u32,
-                weight: Option<&[f64]>,
-                ignore_null: bool,
-                min_periods: u32,
-            ) -> Result<Series> {
-                ChunkWindow::rolling_max(&self.0, window_size, weight, ignore_null, min_periods)
-                    .map(|ca| ca.into_series())
-            }
 
             fn fmt_list(&self) -> String {
                 FmtList::fmt_list(&self.0)
