@@ -7,8 +7,7 @@ use crate::prelude::*;
 use crate::series::implementations::SeriesWrap;
 use crate::series::private::{PrivateSeries, PrivateSeriesNumeric};
 use ahash::RandomState;
-use arrow::array::{ArrayData, ArrayRef};
-use arrow::buffer::Buffer;
+use arrow::array::ArrayRef;
 use std::any::Any;
 use std::borrow::Cow;
 
@@ -32,6 +31,14 @@ impl<T> PrivateSeries for SeriesWrap<ObjectChunked<T>>
 where
     T: PolarsObject,
 {
+    fn agg_first(&self, _groups: &[(u32, Vec<u32>)]) -> Series {
+        self.0.agg_first(_groups)
+    }
+
+    fn agg_last(&self, _groups: &[(u32, Vec<u32>)]) -> Series {
+        self.0.agg_last(_groups)
+    }
+
     fn str_value(&self, index: usize) -> Cow<str> {
         match (&self.0).get(index) {
             None => Cow::Borrowed("null"),
@@ -60,12 +67,13 @@ impl<T> SeriesTrait for SeriesWrap<ObjectChunked<T>>
 where
     T: PolarsObject,
 {
-    fn rename(&mut self, name: &str) {
-        ObjectChunked::rename(&mut self.0, name)
+    #[cfg(feature = "interpolate")]
+    fn interpolate(&self) -> Series {
+        self.0.interpolate().into_series()
     }
 
-    fn array_data(&self) -> Vec<&ArrayData> {
-        ObjectChunked::array_data(&self.0)
+    fn rename(&mut self, name: &str) {
+        ObjectChunked::rename(&mut self.0, name)
     }
 
     fn chunk_lengths(&self) -> ChunkIdIter {
@@ -223,10 +231,6 @@ where
         ChunkUnique::is_duplicated(&self.0)
     }
 
-    fn null_bits(&self) -> Vec<(usize, Option<&Buffer>)> {
-        ObjectChunked::null_bits(&self.0).collect()
-    }
-
     fn reverse(&self) -> Series {
         ChunkReverse::reverse(&self.0).into_series()
     }
@@ -235,8 +239,8 @@ where
         ChunkShift::shift(&self.0, periods).into_series()
     }
 
-    fn fill_none(&self, strategy: FillNoneStrategy) -> Result<Series> {
-        ChunkFillNone::fill_none(&self.0, strategy).map(|ca| ca.into_series())
+    fn fill_null(&self, strategy: FillNullStrategy) -> Result<Series> {
+        ChunkFillNull::fill_null(&self.0, strategy).map(|ca| ca.into_series())
     }
 
     fn fmt_list(&self) -> String {
