@@ -1,22 +1,24 @@
 //! Type agnostic columnar data structure.
 pub use crate::prelude::ChunkCompare;
 use crate::prelude::*;
-use arrow::{array::ArrayRef, buffer::Buffer};
+use arrow::array::ArrayRef;
 pub(crate) mod arithmetic;
 mod comparison;
 pub mod implementations;
 pub(crate) mod iterator;
+pub mod ops;
 
 #[cfg(feature = "object")]
 use crate::chunked_array::object::PolarsObjectSafe;
 use crate::chunked_array::{builder::get_list_builder, ChunkIdIter};
+#[cfg(feature = "rank")]
+use crate::prelude::unique::rank::{rank, RankMethod};
 #[cfg(feature = "groupby_list")]
 use crate::utils::Wrap;
 use crate::utils::{split_ca, split_series};
 use crate::{series::arithmetic::coerce_lhs_rhs, POOL};
 #[cfg(feature = "groupby_list")]
 use ahash::RandomState;
-use arrow::array::ArrayData;
 use arrow::compute::cast;
 use itertools::Itertools;
 use num::NumCast;
@@ -30,6 +32,10 @@ use std::ops::Deref;
 use std::sync::Arc;
 
 pub trait IntoSeries {
+    fn is_series() -> bool {
+        false
+    }
+
     fn into_series(self) -> Series
     where
         Self: Sized;
@@ -58,6 +64,77 @@ pub(crate) mod private {
     }
 
     pub trait PrivateSeries {
+        fn explode_by_offsets(&self, _offsets: &[i64]) -> Series {
+            unimplemented!()
+        }
+
+        /// Apply a rolling mean to a Series. See:
+        /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_mean).
+        #[cfg(feature = "rolling_window")]
+        fn _rolling_mean(
+            &self,
+            _window_size: u32,
+            _weight: Option<&[f64]>,
+            _ignore_null: bool,
+            _min_periods: u32,
+        ) -> Result<Series> {
+            unimplemented!()
+        }
+        /// Apply a rolling sum to a Series. See:
+        /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_sum).
+        #[cfg(feature = "rolling_window")]
+        fn _rolling_sum(
+            &self,
+            _window_size: u32,
+            _weight: Option<&[f64]>,
+            _ignore_null: bool,
+            _min_periods: u32,
+        ) -> Result<Series> {
+            unimplemented!()
+        }
+        /// Apply a rolling min to a Series. See:
+        /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_min).
+        #[cfg(feature = "rolling_window")]
+        fn _rolling_min(
+            &self,
+            _window_size: u32,
+            _weight: Option<&[f64]>,
+            _ignore_null: bool,
+            _min_periods: u32,
+        ) -> Result<Series> {
+            unimplemented!()
+        }
+        /// Apply a rolling max to a Series. See:
+        /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_max).
+        #[cfg(feature = "rolling_window")]
+        fn _rolling_max(
+            &self,
+            _window_size: u32,
+            _weight: Option<&[f64]>,
+            _ignore_null: bool,
+            _min_periods: u32,
+        ) -> Result<Series> {
+            unimplemented!()
+        }
+
+        /// Get an array with the cumulative max computed at every element
+        #[cfg(feature = "cum_agg")]
+        fn _cummax(&self, _reverse: bool) -> Series {
+            panic!("operation cummax not supported for this dtype")
+        }
+
+        /// Get an array with the cumulative min computed at every element
+        #[cfg(feature = "cum_agg")]
+        fn _cummin(&self, _reverse: bool) -> Series {
+            panic!("operation cummin not supported for this dtype")
+        }
+
+        /// Get an array with the cumulative sum computed at every element
+        #[cfg(feature = "cum_agg")]
+        fn _cumsum(&self, _reverse: bool) -> Series {
+            panic!("operation cumsum not supported for this dtype")
+        }
+
         #[cfg(feature = "asof_join")]
         fn join_asof(&self, _other: &Series) -> Result<Vec<Option<u32>>> {
             unimplemented!()
@@ -90,22 +167,22 @@ pub(crate) mod private {
             unimplemented!()
         }
         fn agg_mean(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_min(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_max(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_sum(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_std(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_var(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_first(&self, _groups: &[(u32, Vec<u32>)]) -> Series {
             unimplemented!()
@@ -114,20 +191,20 @@ pub(crate) mod private {
             unimplemented!()
         }
         fn agg_n_unique(&self, _groups: &[(u32, Vec<u32>)]) -> Option<UInt32Chunked> {
-            unimplemented!()
+            None
         }
         fn agg_list(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_quantile(&self, _groups: &[(u32, Vec<u32>)], _quantile: f64) -> Option<Series> {
-            unimplemented!()
+            None
         }
         fn agg_median(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         #[cfg(feature = "lazy")]
         fn agg_valid_count(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            unimplemented!()
+            None
         }
         #[cfg(feature = "pivot")]
         fn pivot<'a>(
@@ -204,29 +281,12 @@ pub(crate) mod private {
 pub trait SeriesTrait:
     Send + Sync + private::PrivateSeries + private::PrivateSeriesNumeric
 {
-    /// Get an array with the cumulative max computed at every element
-    fn cum_max(&self, _reverse: bool) -> Series {
-        panic!("operation cum_max not supported for this dtype")
-    }
-
-    /// Get an array with the cumulative min computed at every element
-    fn cum_min(&self, _reverse: bool) -> Series {
-        panic!("operation cum_min not supported for this dtype")
-    }
-
-    /// Get an array with the cumulative sum computed at every element
-    fn cum_sum(&self, _reverse: bool) -> Series {
-        panic!("operation cum_sum not supported for this dtype")
-    }
+    #[cfg(feature = "interpolate")]
+    fn interpolate(&self) -> Series;
 
     /// Rename the Series.
     fn rename(&mut self, name: &str);
 
-    /// Get Arrow ArrayData
-    fn array_data(&self) -> Vec<&ArrayData> {
-        unimplemented!()
-    }
-
     /// Get the lengths of the underlying chunks
     fn chunk_lengths(&self) -> ChunkIdIter {
         unimplemented!()
@@ -677,11 +737,6 @@ pub trait SeriesTrait:
         unimplemented!()
     }
 
-    /// Get the bits that represent the null values of the underlying ChunkedArray
-    fn null_bits(&self) -> Vec<(usize, Option<&Buffer>)> {
-        unimplemented!()
-    }
-
     /// return a Series in reversed order
     fn reverse(&self) -> Series {
         unimplemented!()
@@ -733,7 +788,7 @@ pub trait SeriesTrait:
     /// * Max fill (replace None with the maximum of the whole array)
     ///
     /// *NOTE: If you want to fill the Nones with a value use the
-    /// [`fill_none` operation on `ChunkedArray<T>`](../chunked_array/ops/trait.ChunkFillNone.html)*.
+    /// [`fill_null` operation on `ChunkedArray<T>`](../chunked_array/ops/trait.ChunkFillNull.html)*.
     ///
     /// # Example
     ///
@@ -742,26 +797,26 @@ pub trait SeriesTrait:
     /// fn example() -> Result<()> {
     ///     let s = Series::new("some_missing", &[Some(1), None, Some(2)]);
     ///
-    ///     let filled = s.fill_none(FillNoneStrategy::Forward)?;
+    ///     let filled = s.fill_null(FillNullStrategy::Forward)?;
     ///     assert_eq!(Vec::from(filled.i32()?), &[Some(1), Some(1), Some(2)]);
     ///
-    ///     let filled = s.fill_none(FillNoneStrategy::Backward)?;
+    ///     let filled = s.fill_null(FillNullStrategy::Backward)?;
     ///     assert_eq!(Vec::from(filled.i32()?), &[Some(1), Some(2), Some(2)]);
     ///
-    ///     let filled = s.fill_none(FillNoneStrategy::Min)?;
+    ///     let filled = s.fill_null(FillNullStrategy::Min)?;
     ///     assert_eq!(Vec::from(filled.i32()?), &[Some(1), Some(1), Some(2)]);
     ///
-    ///     let filled = s.fill_none(FillNoneStrategy::Max)?;
+    ///     let filled = s.fill_null(FillNullStrategy::Max)?;
     ///     assert_eq!(Vec::from(filled.i32()?), &[Some(1), Some(2), Some(2)]);
     ///
-    ///     let filled = s.fill_none(FillNoneStrategy::Mean)?;
+    ///     let filled = s.fill_null(FillNullStrategy::Mean)?;
     ///     assert_eq!(Vec::from(filled.i32()?), &[Some(1), Some(1), Some(2)]);
     ///
     ///     Ok(())
     /// }
     /// example();
     /// ```
-    fn fill_none(&self, _strategy: FillNoneStrategy) -> Result<Series> {
+    fn fill_null(&self, _strategy: FillNullStrategy) -> Result<Series> {
         unimplemented!()
     }
 
@@ -797,50 +852,6 @@ pub trait SeriesTrait:
     fn quantile_as_series(&self, _quantile: f64) -> Result<Series> {
         unimplemented!()
     }
-    /// Apply a rolling mean to a Series. See:
-    /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_mean).
-    fn rolling_mean(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Series> {
-        unimplemented!()
-    }
-    /// Apply a rolling sum to a Series. See:
-    /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_sum).
-    fn rolling_sum(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Series> {
-        unimplemented!()
-    }
-    /// Apply a rolling min to a Series. See:
-    /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_min).
-    fn rolling_min(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Series> {
-        unimplemented!()
-    }
-    /// Apply a rolling max to a Series. See:
-    /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_max).
-    fn rolling_max(
-        &self,
-        _window_size: u32,
-        _weight: Option<&[f64]>,
-        _ignore_null: bool,
-        _min_periods: u32,
-    ) -> Result<Series> {
-        unimplemented!()
-    }
 
     fn fmt_list(&self) -> String {
         "fmt implemented".into()
@@ -1066,9 +1077,17 @@ pub trait SeriesTrait:
 
     #[cfg(feature = "mode")]
     #[cfg_attr(docsrs, doc(cfg(feature = "mode")))]
+    /// Compute the most occurring element in the array.
     fn mode(&self) -> Result<Series> {
         unimplemented!()
     }
+
+    #[cfg(feature = "rolling_window")]
+    /// Apply a custom function over a rolling/ moving window of the array.
+    /// This has quite some dynamic dispatch, so prefer rolling_min, max, mean, sum over this.
+    fn rolling_apply(&self, _window_size: usize, _f: &dyn Fn(&Series) -> Series) -> Result<Series> {
+        panic!("rolling apply not implemented for this dtype. Only implemented for numeric data.")
+    }
 }
 
 impl<'a> (dyn SeriesTrait + 'a) {
@@ -1325,13 +1344,7 @@ impl Series {
         match self.dtype() {
             DataType::Float32 => Ok(self.f32().unwrap().is_nan()),
             DataType::Float64 => Ok(self.f64().unwrap().is_nan()),
-            _ => Err(PolarsError::InvalidOperation(
-                format!(
-                    "is_nan not supported for series with dtype {:?}",
-                    self.dtype()
-                )
-                .into(),
-            )),
+            _ => Ok(BooleanChunked::full(self.name(), false, self.len())),
         }
     }
 
@@ -1340,13 +1353,7 @@ impl Series {
         match self.dtype() {
             DataType::Float32 => Ok(self.f32().unwrap().is_not_nan()),
             DataType::Float64 => Ok(self.f64().unwrap().is_not_nan()),
-            _ => Err(PolarsError::InvalidOperation(
-                format!(
-                    "is_nan not supported for series with dtype {:?}",
-                    self.dtype()
-                )
-                .into(),
-            )),
+            _ => Ok(BooleanChunked::full(self.name(), true, self.len())),
         }
     }
 
@@ -1526,15 +1533,162 @@ impl Series {
     }
 
     #[cfg(feature = "dot_product")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "dot_product")))]
     pub fn dot(&self, other: &Series) -> Option<f64> {
         (self * other).sum::<f64>()
     }
 
     #[cfg(feature = "row_hash")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "row_hash")))]
     /// Get a hash of this Series
     pub fn hash(&self, build_hasher: ahash::RandomState) -> UInt64Chunked {
         UInt64Chunked::new_from_aligned_vec(self.name(), self.0.vec_hash(build_hasher))
     }
+
+    /// Get an array with the cumulative max computed at every element
+    #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
+    pub fn cummax(&self, _reverse: bool) -> Series {
+        #[cfg(feature = "cum_agg")]
+        {
+            self._cummax(_reverse)
+        }
+        #[cfg(not(feature = "cum_agg"))]
+        {
+            panic!("activate 'cum_agg' feature")
+        }
+    }
+
+    /// Get an array with the cumulative min computed at every element
+    #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
+    pub fn cummin(&self, _reverse: bool) -> Series {
+        #[cfg(feature = "cum_agg")]
+        {
+            self._cummin(_reverse)
+        }
+        #[cfg(not(feature = "cum_agg"))]
+        {
+            panic!("activate 'cum_agg' feature")
+        }
+    }
+
+    /// Get an array with the cumulative sum computed at every element
+    #[cfg_attr(docsrs, doc(cfg(feature = "cum_agg")))]
+    pub fn cumsum(&self, _reverse: bool) -> Series {
+        #[cfg(feature = "cum_agg")]
+        {
+            match self.dtype() {
+                DataType::Boolean => self
+                    .cast_with_dtype(&DataType::UInt32)
+                    .unwrap()
+                    ._cumsum(_reverse),
+                _ => self._cumsum(_reverse),
+            }
+        }
+        #[cfg(not(feature = "cum_agg"))]
+        {
+            panic!("activate 'cum_agg' feature")
+        }
+    }
+
+    /// Apply a rolling mean to a Series. See:
+    /// [ChunkedArray::rolling_mean](crate::prelude::ChunkWindow::rolling_mean).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    pub fn rolling_mean(
+        &self,
+        _window_size: u32,
+        _weight: Option<&[f64]>,
+        _ignore_null: bool,
+        _min_periods: u32,
+    ) -> Result<Series> {
+        #[cfg(feature = "rolling_window")]
+        {
+            self._rolling_mean(_window_size, _weight, _ignore_null, _min_periods)
+        }
+        #[cfg(not(feature = "rolling_window"))]
+        {
+            panic!("activate 'rolling_window' feature")
+        }
+    }
+    /// Apply a rolling sum to a Series. See:
+    /// [ChunkedArray::rolling_sum](crate::prelude::ChunkWindow::rolling_sum).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    pub fn rolling_sum(
+        &self,
+        _window_size: u32,
+        _weight: Option<&[f64]>,
+        _ignore_null: bool,
+        _min_periods: u32,
+    ) -> Result<Series> {
+        #[cfg(feature = "rolling_window")]
+        {
+            self._rolling_sum(_window_size, _weight, _ignore_null, _min_periods)
+        }
+        #[cfg(not(feature = "rolling_window"))]
+        {
+            panic!("activate 'rolling_window' feature")
+        }
+    }
+    /// Apply a rolling min to a Series. See:
+    /// [ChunkedArray::rolling_min](crate::prelude::ChunkWindow::rolling_min).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    pub fn rolling_min(
+        &self,
+        _window_size: u32,
+        _weight: Option<&[f64]>,
+        _ignore_null: bool,
+        _min_periods: u32,
+    ) -> Result<Series> {
+        #[cfg(feature = "rolling_window")]
+        {
+            self._rolling_min(_window_size, _weight, _ignore_null, _min_periods)
+        }
+        #[cfg(not(feature = "rolling_window"))]
+        {
+            panic!("activate 'rolling_window' feature")
+        }
+    }
+    /// Apply a rolling max to a Series. See:
+    /// [ChunkedArray::rolling_max](crate::prelude::ChunkWindow::rolling_max).
+    #[cfg_attr(docsrs, doc(cfg(feature = "rolling_window")))]
+    pub fn rolling_max(
+        &self,
+        _window_size: u32,
+        _weight: Option<&[f64]>,
+        _ignore_null: bool,
+        _min_periods: u32,
+    ) -> Result<Series> {
+        #[cfg(feature = "rolling_window")]
+        {
+            self._rolling_max(_window_size, _weight, _ignore_null, _min_periods)
+        }
+        #[cfg(not(feature = "rolling_window"))]
+        {
+            panic!("activate 'rolling_window' feature")
+        }
+    }
+
+    #[cfg(feature = "rank")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "rank")))]
+    pub fn rank(&self, method: RankMethod) -> Series {
+        rank(self, method)
+    }
+
+    /// Cast throws an error if conversion had overflows
+    pub fn strict_cast(&self, data_type: &DataType) -> Result<Series> {
+        let s = self.cast_with_dtype(data_type)?;
+        if self.null_count() != s.null_count() {
+            Err(PolarsError::ComputeError(
+                format!(
+                    "strict conversion of cast from {:?} to {:?} failed. consider non-strict cast.",
+                    self.dtype(),
+                    data_type
+                )
+                .into(),
+            ))
+        } else {
+            Ok(s)
+        }
+    }
 }
 
 impl Deref for Series {
@@ -1654,14 +1808,22 @@ impl std::convert::TryFrom<(&str, Vec<ArrayRef>)> for Series {
             ArrowDataType::Utf8 => {
                 let chunks = chunks
                     .iter()
-                    .map(|arr| cast(arr, &ArrowDataType::LargeUtf8).unwrap())
+                    .map(|arr| {
+                        cast::cast(arr.as_ref(), &ArrowDataType::LargeUtf8)
+                            .unwrap()
+                            .into()
+                    })
                     .collect_vec();
                 Ok(Utf8Chunked::new_from_chunks(name, chunks).into_series())
             }
             ArrowDataType::List(fld) => {
                 let chunks = chunks
                     .iter()
-                    .map(|arr| cast(arr, &ArrowDataType::LargeList(fld.clone())).unwrap())
+                    .map(|arr| {
+                        cast::cast(arr.as_ref(), &ArrowDataType::LargeList(fld.clone()))
+                            .unwrap()
+                            .into()
+                    })
                     .collect();
                 Ok(ListChunked::new_from_chunks(name, chunks).into_series())
             }
@@ -1718,7 +1880,11 @@ impl std::convert::TryFrom<(&str, Vec<ArrayRef>)> for Series {
             ArrowDataType::Timestamp(TimeUnit::Millisecond, None) => {
                 let chunks = chunks
                     .iter()
-                    .map(|arr| cast(arr, &ArrowDataType::Date64).unwrap())
+                    .map(|arr| {
+                        cast::cast(arr.as_ref(), &ArrowDataType::Date64)
+                            .unwrap()
+                            .into()
+                    })
                     .collect();
                 Ok(Date64Chunked::new_from_chunks(name, chunks).into_series())
             }
@@ -1738,6 +1904,15 @@ impl TryFrom<(&str, ArrayRef)> for Series {
     }
 }
 
+impl TryFrom<(&str, Box<dyn Array>)> for Series {
+    type Error = PolarsError;
+
+    fn try_from(name_arr: (&str, Box<dyn Array>)) -> Result<Self> {
+        let (name, arr) = name_arr;
+        Series::try_from((name, vec![arr.into()]))
+    }
+}
+
 impl Default for Series {
     fn default() -> Self {
         Int64Chunked::default().into_series()
@@ -1761,6 +1936,10 @@ impl IntoSeries for Arc<dyn SeriesTrait> {
 }
 
 impl IntoSeries for Series {
+    fn is_series() -> bool {
+        true
+    }
+
     fn into_series(self) -> Series {
         self
     }
@@ -1790,7 +1969,6 @@ where
 mod test {
     use crate::prelude::*;
     use crate::series::*;
-    use arrow::array::*;
 
     #[test]
     fn cast() {
@@ -1813,7 +1991,7 @@ mod test {
 
     #[test]
     fn new_series_from_arrow_primitive_array() {
-        let array = UInt32Array::from(vec![1, 2, 3, 4, 5]);
+        let array = UInt32Array::from_slice(&[1, 2, 3, 4, 5]);
         let array_ref: ArrayRef = Arc::new(array);
 
         Series::try_from(("foo", array_ref)).unwrap();
